# -*- coding: utf-8 -*-

import logging
import math
from itertools import combinations

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from pymatgen.analysis.local_env import CrystalNN
from scipy.interpolate import RegularGridInterpolator
from scipy.signal import savgol_filter
from scipy.spatial import ConvexHull
from tqdm import tqdm

from simmate.apps.badelf.core.grid import Grid
from simmate.toolkit import Structure


class PartitioningToolkit:
    """
    A set of tools for aiding in partitioning a 3D voxel grid. These tools make
    up the basis for the partitioning utilized in BadELF and the related
    VoronELF algorithms.

    Args:
        grid (Grid): A BadELF app Grid type object.
    """

    def __init__(self, grid: Grid):
        self.grid = grid.copy()

    def get_partitioning_line_from_voxels(
        self,
        site_voxel_coord: ArrayLike | list,
        neigh_voxel_coord: ArrayLike | list,
        # method: str = "linear",
        method: str = "linear",
    ):
        """
        Finds a line of voxel positions between two atom sites and then finds the value
        of the partitioning grid at each of these positions. The values are found
        using an interpolation function defined using SciPy's RegularGridInterpeter.

        Args:
            site_voxel_coord (ArrayLike): The voxel coordinates of an atomic site
            neigh_voxel_coord (ArrayLike): The voxel coordinates of a neighboring
                site
            method (str): The method of interpolation. 'cubic' is more rigorous
                than 'linear'

        Results:
            A list with 200 pairs of voxel coordinates and data values along
            a line between two positions.
        """
        grid_data = self.grid.copy().total
        steps = 200
        slope = [b - a for a, b in zip(site_voxel_coord, neigh_voxel_coord)]
        slope_increment = [float(x) / steps for x in slope]

        # get a list of points along the connecting line. First add the original
        # site
        position = site_voxel_coord
        line = [
            [
                round((float(((a - 1) % b) + 1)), 12)
                for a, b in zip(position, grid_data.shape)
            ]
        ]
        for i in range(steps):
            # move position by slope_increment
            position = [float(a + b) for a, b in zip(position, slope_increment)]

            # Wrap values back into cell
            # We must do (a-1) to shift the voxel index (1 to grid_max+1) onto a
            # normal grid, (0 to grid_max), then do the wrapping function (%), then
            # shift back onto the VASP voxel index.
            position = [
                round((float(((a - 1) % b) + 1)), 12)
                for a, b in zip(position, grid_data.shape)
            ]

            line.append(position)

        # The partitioning uses a padded grid and grid interpolation to find the
        # location of dividing planes.
        padded_grid_data = np.pad(grid_data, 1, mode="wrap")

        # interpolate grid to find values that lie between voxels. This is done
        # with a cruder interpolation here and then the area close to the minimum
        # is examened more closely with a more rigorous interpolation in
        # get_line_frac_min
        a, b, c = self.grid.get_grid_axes(1)
        fn = RegularGridInterpolator((a, b, c), padded_grid_data, method=method)
        # get a list of the ELF values along the line
        values = []

        for pos in line:
            adjusted_pos = [x for x in pos]
            value = float(fn(adjusted_pos))
            values.append(value)

        # smooth line with savgol filter
        # values = savgol_filter(values, 20, 3)
        return line, values

    def get_partitioning_line_from_indices(self, i: int, j: int):
        """
        Gets the voxel positions and elf values for points between two sites in
        the structure.

        Parameters:
        - i: index of first site in the structure
        - j: index of second site in the structure

        Returns:
        - Two lists, one of positions in voxel coordinates and another of elf
        values
        """
        grid = self.grid.copy()
        site_voxel_coord = grid.get_voxel_coords_from_index(i)
        neigh_voxel_coord = grid.get_voxel_coords_from_index(j)
        return self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord
        )

    def get_partitioning_line_from_cart_coords(
        self, site_cart_coords, neigh_cart_coords
    ):
        """
        Gets the voxel positions and elf values for points between two sites in
        the structure given as cartesian coordinates. This method can also be
        used to find the values in the ELF between two arbitrary points in the
        structure.

        Parameters:
        - site_cart_coords: cartesian coordinates of a site in the structure
        - neigh_cart_coords: cartesian coordinates of a second site in the structure

        Returns:
        - Two lists, one of positions in voxel coordinates and another of elf
        values
        """
        grid = self.grid.copy()
        site_voxel_coord = grid.get_voxel_coords_from_cart(site_cart_coords)
        neigh_voxel_coord = grid.get_voxel_coords_from_cart(neigh_cart_coords)
        return self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord
        )

    @staticmethod
    def _check_partitioning_line_for_symmetry(values: list, tolerance: float = 0.1):
        """
        Check if the values are roughly symmetric.

        Parameters:
        - values: List of numeric values
        - tolerance: Tolerance level for symmetry check

        Returns:
        - True if roughly symmetric, False otherwise
        """
        n = len(values)

        # Check if the list has an even number of elements
        if n % 2 != 0:
            # remove the center if odd number
            center_index = math.ceil(n / 2)
            values.pop(center_index)

        # Split the list into two halves
        half_size = n // 2
        first_half = values[:half_size]
        second_half = values[half_size:]

        # Reverse the second half
        reversed_second_half = list(reversed(second_half))

        # Check if the values are roughly equal within the given tolerance
        for val1, val2 in zip(first_half, reversed_second_half):
            if abs(val1 - val2) > tolerance:
                return False

        return True

    @staticmethod
    def find_minimum(values: list | ArrayLike):
        """
        Finds the local minima in a list of values and returns the index and value
        at each as a list of form [[min_index1, min_value1], [min_index2, min_value2], ...]

        Args:
            values (list): The list of values to find the minima of

        results:
            A list of minima represented by [index, value]
        """
        minima = [
            [i, y]
            for i, y in enumerate(values)
            if ((i == 0) or (values[i - 1] >= y))
            and ((i == len(values) - 1) or (y < values[i + 1]))
        ]
        return minima

    @staticmethod
    def find_maximum(values: list | ArrayLike):
        """
        Finds the local maxima in a list of values and returns the index and value
        at each as a list of form [[max_index1, max_value1], [max_index2, max_value2], ...]

        Args:
            values (list): The list of values to find the minima of

        results:
            A list of maxima represented by [index, value]
        """
        maxima = [
            [i, y]
            for i, y in enumerate(values)
            if ((i == 0) or (values[i - 1] <= y))
            and ((i == len(values) - 1) or (y > values[i + 1]))
        ]
        return maxima

    @staticmethod
    def get_closest_extrema_to_center(
        values: list | ArrayLike,
        extrema: list | ArrayLike,
    ):
        """
        Takes a list of values and the relative extrema (either minima or maxima)
        and finds which extrema is closest to the center of the line.

        Args:
            values (list): A list of values
            extrema (list): A list of extrema of form [index, value]

        results:
            The global extreme of form [index, value]
        """
        midpoint = len(values) / 2
        differences = []
        for pos, val in extrema:
            diff = abs(pos - midpoint)
            differences.append(diff)
        min_pos = differences.index(min(differences))
        global_extrema = extrema[min_pos]
        return global_extrema

    def get_line_minimum_as_frac(
        self,
        positions: list,
        values: list | ArrayLike,
        site_string: str,
        neigh_string: str,
    ):
        """
        Finds the minimum point of a list of values along a line, then returns the
        fractional position of this values position along the line.

        Args:
            positions (list):
                A list of positions given as voxel coordinates along the line
                of interest
            values (list):
                A list of values to find the minimum of
            site_string (str):
                The symbol of the atom at the start of the line
            neigh_string (str):
                The symbol of the atom at the end of the line

        results:
            The global minimum of form [line_position, value, frac_position]
        """

        if site_string == neigh_string:
            list_values = list(values)
            # We have the same type of atom on either side. We want to check
            # if they are the same and if they are return a frac of 0.5. This
            # is because usually there will be some slight covalency between
            # atoms of the same type, but they should share the area equally
            symmetric = self._check_partitioning_line_for_symmetry(list_values)
        else:
            symmetric = False

        # If we found symmetry, we return a global min exactly at the center
        if symmetric:
            # 100 is the index directly at the center of the line
            elf_value = values[100]
            elf_min_frac = 0.5
            global_min = [100, elf_value, elf_min_frac]
        else:
            # We either don't have the same atoms, or the same atoms are not
            # symmetric along the elf.
            # minima function gives all local minima along the values
            minima = self.find_minimum(values)
            # maxima = find_maximum(values)

            # then we grab the local minima closest to the midpoint of the values
            global_min = self.get_closest_extrema_to_center(values, minima)
            global_min = self._refine_line_min_frac(
                positions=positions, elf_min_index=global_min[0]
            )

        return global_min

    def _refine_line_min_frac(self, positions, elf_min_index):
        """
        Refines the location of the minimum along an ELF line between two sites.
        To do this, the initial estimate from a linear interpolation of the line
        is used and a cubic interpolation is used in a smaller area around the
        estimated point. The sampled area is adjusted if it is found to not be
        centered on the new more accurate minimum.

        Args:
            positions (list):
                A list of positions given as voxel coordinates along the line
                of interest
            elf_min_index (int):
                The index along the line at which the linear interpolation estimated
                the minimum.

        results:
            The global minimum of form [line_position, value, frac_position]
        """
        amount_to_pad = 10
        grid = self.grid.copy()
        padded = np.pad(grid.total, amount_to_pad, mode="wrap")

        # interpolate the grid with a more rigorous method to find more exact value
        # for the plane.
        a, b, c = grid.get_grid_axes(10)
        fn = RegularGridInterpolator((a, b, c), padded, method="cubic")

        # create variables for if the line needs to be shifted from what the
        # rough partitioning found
        centered = False
        amount_to_shift = 0
        attempts = 0

        while centered == False:
            if attempts == 5:
                break
            else:
                attempts += 1
                # If the position wasn't centered previously, we need to shift
                # the index
                elf_min_index = elf_min_index + amount_to_shift
                line_section = positions[elf_min_index - 3 : elf_min_index + 4]
                line_section_x = [
                    i for i in range(elf_min_index - 3, elf_min_index + 4)
                ]

                values_fine = []
                # Get the list of values from the interpolated grid
                for pos in line_section:
                    new_pos = [i + amount_to_pad - 1 for i in pos]
                    value_fine = float(fn(new_pos))
                    values_fine.append(value_fine)

                # Find the minimum value of this line as well as the index for this value's
                # position.
                minimum_value = min(values_fine)
                min_pos = values_fine.index(minimum_value)  # + global_min_pos[0]-5

                if min_pos == 4:
                    # Our line is centered and we can move on
                    centered = True
                else:
                    # Our line is not centered and we need to adjust it
                    amount_to_shift = min_pos - 4

        if not centered:
            # The above sometimes fails because the linear fitting gives a guess
            # for the minimum that isn't close. To handle this we treat these
            # situations rigorously
            values = []

            # Get the ELF value for every position in the line.
            for pos in positions:
                new_pos = [i + amount_to_pad - 1 for i in pos]
                value = float(fn(new_pos))
                values.append(value)

            # Get a list of all of the minima along the line
            minima = self.find_minimum(values)

            # then we grab the local minima closest to the midpoint of the line
            global_min = self.get_closest_extrema_to_center(values, minima)

            # now we want a small section of the line surrounding the minimum
            values_fine = values[global_min[0] - 3 : global_min[0] + 4]
            line_section_x = [i for i in range(global_min[0] - 3, global_min[0] + 4)]

        # now that we've found the values surrounding the minimum of our line,
        # we can fit these values to a 2nd degree polynomial and solve for its
        # minimum point
        d, e, f = np.polyfit(line_section_x, values_fine, 2)
        x = -e / (2 * d)
        elf_min_index_new = x
        elf_min_value_new = np.polyval(np.array([d, e, f]), x)
        elf_min_frac_new = elf_min_index_new / (len(positions) - 1)

        return [elf_min_index_new, elf_min_value_new, elf_min_frac_new]

    @staticmethod
    def get_voxel_coords_from_min_along_line(
        global_min_frac: float,
        site_voxel_coords: ArrayLike | list,
        neigh_voxel_coords: ArrayLike | list,
    ):
        """
        Gives the voxel position/coords for the minimum along a partitioning line
        between two atoms. The minimum point is where the dividing plane should lie.

        Args:
            global_min_frac (float): The global minimum fraction
            site_voxel_coords (ArrayLike): The voxel coordinates of the site at
                the start of the line.
            neigh_voxel_coords (ArrayLike): The voxel coordinates of the neighbor
                atom at the end of the line.
        """

        # get the vector that points between the site and its neighbor (in vox  coords)
        difference = [b - a for a, b in zip(site_voxel_coords, neigh_voxel_coords)]
        # get the vector that points from the site to the global_min_frac
        min_pos = [x * global_min_frac for x in difference]
        # add the vector components of the site and minimum point together to get
        # the vector pointing directly to the minimum
        min_pos = [(a + b) for a, b in zip(min_pos, site_voxel_coords)]
        return np.array(min_pos)

    def get_unit_vector(self, site_voxel_coords, neigh_voxel_coords):
        """
        Gets the unit vector pointing in the direction between two atom sites.

        Args:
            site_voxel_coord (ArrayLike): The voxel coordinates of an atomic site
            neigh_voxel_coord (ArrayLike): The voxel coordinates of a neighboring
                site
        """
        grid = self.grid.copy()
        # get the positions of the atoms in cartesian coordinates.
        site_cart_coords = grid.get_cart_coords_from_vox(site_voxel_coords)
        real_neigh_voxel_coords = grid.get_cart_coords_from_vox(neigh_voxel_coords)
        # The equation of a plane passing through (x1,y1,z1) with normal vector
        # [a,b,c] is:
        #     a(x-x1) + b(y-y1) + c(z-z1) = 0
        # we want the normal vector to be unit vector because later on we
        # will use this information to get the distance of a voxel from
        # the plane.
        normal_vector = [
            (b - a) for a, b in zip(site_cart_coords, real_neigh_voxel_coords)
        ]
        unit_vector = [(x / np.linalg.norm(normal_vector)) for x in normal_vector]
        return np.array(unit_vector)

    @staticmethod
    def get_plane_sign(
        point: ArrayLike | list,
        plane_vector: ArrayLike | list,
        plane_point: ArrayLike | list,
    ):
        """
        Gets the sign associated with a point compared with a plane.

        Args:
            point (ArrayLike): A point in cartesian coordinates to compare with
                a plane
            plane_vector (ArrayLike): The vector normal to the plane of interest
            plane_point (ArrayLike): A point on the plane of interest

        results:
            The sign of the point compared with the plane and the distance of
            the point to the plane.
        """
        # get all of the points in cartesian coordinates
        x, y, z = plane_point
        a, b, c = plane_vector
        x1, y1, z1 = point
        value_of_plane_equation = a * (x - x1) + b * (y - y1) + c * (z - z1)
        # get the distance of the point from the plane with some allowance of error.
        if value_of_plane_equation > 1e-6:
            return "positive", abs(value_of_plane_equation)
        elif value_of_plane_equation < -1e-6:
            return "negative", abs(value_of_plane_equation)
        else:
            return "zero", abs(value_of_plane_equation)

    def get_distance_to_min(
        self,
        minimum_point: ArrayLike | list,
        site_voxel_coords: ArrayLike | list,
    ):
        """
        Gets the distance from an atom to the minimum point along partitioning line
        between it and another atom.

        Args:
            minimum_point (ArrayLike): A point in cartesian coordinates where
                there is a minimum along a line
            site_voxel_coords (ArrayLike): The voxel coordinates of a site that
                is at one end of the line that the minimum belongs to

        results:
            The distance from the site to the minimum
        """
        grid = self.grid.copy()
        site_cart_coords = grid.get_cart_coords_from_vox(site_voxel_coords)
        distance = sum(
            [(b - a) ** 2 for a, b in zip(site_cart_coords, minimum_point)]
        ) ** (0.5)
        return distance

    def get_elf_ionic_radius(
        self,
        site_index: int,
        method: str = "linear",
        structure: Structure = None,
        site_is_electride: bool = False,
    ):
        """
        This method gets the ELF ionic radius. It interpolates the ELF values
        between a site and it's closest neighbor and returns the distance between
        the atom and the minimum in this line. This has been shown to be very
        similar to the Shannon Crystal Radius, but gives more specific values

        Args:
            site_index (int):
                An integer value referencing an atom in the structure
            method (str):
                Whether to use linear or cubic interpolation
            structure (Structure):
                The structure to use if it is not the same as the one stored
                in the PartitioningToolkit instance
            site_is_electride (bool):
                Whether the site in question is an electride site.
        """
        if structure is None:
            structure = self.grid.structure.copy()
        # get closest neighbor for the given site
        neighbors = structure.get_neighbors(structure[site_index], 15)

        # Get the closest neighbor to the site. We put the nearest neighbors in
        # a dataframe, sort by distance, and then select the nearest one.
        site_df = pd.DataFrame(columns=["neighbor", "distance"])
        for neigh in neighbors:
            # get distance
            distance = math.dist(neigh.coords, structure[site_index].coords)
            # add neighbor, distance pair to df
            site_df.loc[len(site_df)] = [neigh, distance]
        # sort by distance and truncate to first 50
        site_df = site_df.sort_values(by="distance")
        for i, row in site_df.iterrows():
            neigh = row["neighbor"]
            if neigh.species_string != "He":
                closest_neighbor = neigh
                bond_dist = row["distance"]
                break
            else:
                continue

        # Get site voxel coords from cartesian coordinates
        site_voxel_coord = self.grid.get_voxel_coords_from_cart(
            structure[site_index].coords
        )
        # Get neighbor voxel coords and then the line between the sites/voxels
        neigh_voxel_coord = self.grid.get_voxel_coords_from_cart(
            closest_neighbor.coords
        )
        elf_positions, elf_values = self.get_partitioning_line_from_voxels(
            site_voxel_coord,
            neigh_voxel_coord,
            method,
        )

        # Get site and neighbor strings
        site_string = self.grid.structure[site_index].species_string
        neighbor_string = closest_neighbor.species_string

        max_bond_dist = (
            structure[site_index].specie.atomic_radius
            + closest_neighbor.specie.atomic_radius
        )
        # If the bond is larger than the some of both atomic radii it is likely
        # that there is an electride between the atom and its closest neighbor.
        # We want to get the distance to the minimum closer to the atom than
        # its neighbor.
        if bond_dist > max_bond_dist:
            # We get all of the minima along  the line. We then find the distance
            # of each minima from the middle of the line, remove any that are
            # closer to the neighboring atom, and find the one closest to the
            # center of the bond.
            minima = self.find_minimum(elf_values)
            mid_point = len(elf_values) / 2
            dists_to_min = np.array([mid_point - x[0] for x in minima])
            dists_to_min = dists_to_min[dists_to_min >= 0]
            min_pos = minima[np.argmin(dists_to_min)]
            # Append the fractional distance to the minimum of the line.
            min_pos.append(min_pos[0] / (len(elf_values) - 1))
        else:
            # Get the min position along the line
            min_pos = self.get_line_minimum_as_frac(
                elf_positions, elf_values, site_string, neighbor_string
            )
        min_voxel_coord = self.get_voxel_coords_from_min_along_line(
            min_pos[2], site_voxel_coord, neigh_voxel_coord
        )
        min_cart_coord = self.grid.get_cart_coords_from_vox(min_voxel_coord)
        distance_to_min = self.get_distance_to_min(min_cart_coord, site_voxel_coord)

        return distance_to_min

    def get_closest_neighbors(
        self,
        structure: Structure = None,
    ):
        """
        Function for getting the closest neighbors to an atom. Uses the CrystalNN
        class from pymatgen. This is intended to help quickly check for covalency
        in the structure and may eventually be removed.

        Results:
            A dictionary relating atomic sites to pymatgen neighbor objects
        """
        if structure is None:
            structure = self.grid.structure
        c = CrystalNN(search_cutoff=5)
        closest_neighbors = {}
        for i in range(len(structure)):
            _, _, d = c.get_nn_data(structure, n=i)
            biggest = max(d)
            closest_neighbors[i] = d[biggest]
        return closest_neighbors

    def get_set_number_of_neighbors(self, neighbor_num: int = 26):
        """
        Function for getting the closest neighbors. This is necessary for partitioning
        because the CrystalNN function from pymatgen will not always find the
        correct set of atoms needed to create a full partitioning set. We default
        to 26, but if this is not large enough we automatically increase the amount.

        Args:
            neighbor_num (int): The number of nearest neighbors to find

        Results:
            A list of relating an atoms index to its  neighbors.

        """
        structure = self.grid.structure.copy()

        # Get all possible neighbor atoms for each atom within 15 angstroms
        all_neighbors = structure.get_all_neighbors(15)
        neighbors = []

        for site, neighs in enumerate(all_neighbors):
            # Get the coordinates for each site/neighbor pair and convert to numpy array
            site_coords = []
            neigh_coords = []
            site_coord = structure[site].coords
            for neigh in neighs:
                neigh_coord = neigh.coords
                site_coords.append(site_coord)
                neigh_coords.append(neigh_coord)
            site_coords = np.array(site_coords)
            neigh_coords = np.array(neigh_coords)
            # Calculate the distance between the site and each neighbor
            distances = np.linalg.norm(neigh_coords - site_coords, axis=1)

            # Create a dataframe of neighbors and distance
            site_df = pd.DataFrame(columns=["neighbors", "distances"])
            site_df["neighbors"] = neighs
            site_df["distances"] = distances
            # sort by distance and truncate to requested length
            site_df = site_df.sort_values(by="distances")
            site_df = site_df.iloc[0:neighbor_num]
            # Append neighbors to our neighbor list
            neighbors.append(site_df["neighbors"].to_list())

        return neighbors

    @classmethod
    def check_bond_for_covalency(
        cls,
        values: list,
    ):
        """
        Checks for covalent/metallic behavior along a bond. This is done by comparing
        the closest local maximum and minimum to the center of the bond. If the
        maximum is closer, the bond is considered to have some covalent behavior.

        Args:
            values (list): The list of values to check along

        Returns:
            True if the bond has covalent behavior and False if not.
        """
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # In the future, these covalent areas should also be assigned sites so that
        # covalent electrides can be handled as well. There is also probably a more
        # rigorous point to compare to than the center of the bond depending on the
        # sizes of the atoms. Maybe compare the ratio of atomic radii and set the
        # comparison point further from the larger atom based on this ratio?
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        # get a smoothed elf line
        values = savgol_filter(values, 20, 3)

        # get the center of the ELF line
        midpoint = len(values) / 2

        # minima function gives all local minima along the values. maxima does the reverse
        minima = cls.find_minimum(values)
        maxima = cls.find_maximum(values)

        # then we grab the local minima closest to the midpoint of the values
        global_min = cls.get_closest_extrema_to_center(values, minima)
        global_max = cls.get_closest_extrema_to_center(values, maxima)

        # get the distance from the minimum and maximum to the center. The position
        # of the extrema is stored in the first index of the extrema list
        min_dist = abs(midpoint - global_min[0])
        max_dist = abs(midpoint - global_max[0])

        # get the ELF values at the minimum and maximum. This is stored in the
        # second index of the extrema list
        min_elf = global_min[1]
        max_elf = global_max[1]

        # If the maximum is closer to the center of the line than the minimum, then
        # we consider this bond to have metallic or covalent behavior and return True
        if min_dist > max_dist:
            return True

        # If the minimum is closer we need to do an extra check. If the minimum has
        # a value that is over half the value of the maximum, it is likely that there
        # are two local maxima that are close to the center of the bond, which we
        # also view as covalent and return as True. If it is less than half the
        # value of the maximum, this bond is not strongly covalent and the partitioning
        # will work, so we return False.
        if min_dist <= max_dist:
            if min_elf > max_elf / 2:
                return True
            else:
                return False

    def check_closest_neighbor_for_same_type(
        self,
        closest_neighbors: dict,
    ):
        """
        This function checks indirectly for covalency by checking if an atom's
        closest neighbor is an atom of the same type. This suggests that the
        atoms are covalently bonded. This is to account for the fact that in the
        check_structure_for_covalency method, we skip any atom pairs that are the
        same since they are likely to return as covalent regardless of how far apart
        they are.

        Args:
            closest_neighbors (dict): A dictionary relating sites to pymatgen
                Neigh objects determined from CrystalNN
        Returns:
            Nothing
        """
        structure = self.grid.structure
        # we initially assume that the closest atom is not one of the same type.
        same_atom_close = False
        for site_index, neighs in closest_neighbors.items():
            # get site atom string
            specie = structure[site_index].species_string
            # if the species is He, it is a dummy atom in an electride site and we
            # want to skip it
            if specie == "He":
                continue

            # to check if the closest atom is the same type of atom, we need ot find
            # which atom is closest and which of the same atom is closest. First we
            # make lists for all nearby atom distances and same atom distances.
            site_distances = []
            same_atom_distances = []

            # Now we find the sites coordinates and loop over each of its neighbors
            # to find their coords. For each we find the distance away and add the
            # distance to our distances list. If they are the same atom type we also
            # add them to the same_atom_distances list.
            site_coords = structure[site_index].coords
            for neigh in neighs:
                neigh_coords = neigh["site"].coords
                site_distances.append(math.dist(site_coords, neigh_coords))
                # check that the neighbors are the same species
                if neigh["site"].specie == specie:
                    same_atom_distances.append(math.dist(site_coords, neigh_coords))
            closest_atom_dist = min(site_distances)
            # find the closest atom of the same type. Sometimes this doesn't exist
            # in the nearest neighbors set so we set this to a large distance.
            try:
                closest_same_atom_dist = min(same_atom_distances)
            except:
                closest_same_atom_dist = 10
            # if the closest atom is an atom of the same time, we
            if closest_same_atom_dist <= closest_atom_dist:
                same_atom_close = True
                break
        # if the same atom is found as the closest atom, we want to raise an error
        if same_atom_close:
            raise Exception(
                """
                An atom's closest neighbor was found to be an atom of the same type.
                This very likely indicates that these atoms are bonded covalently.
                Unfortunately, the current version of BadELF does not have a way to 
                partition peaks in the ELF from covalency, though this will hopefully 
                be implemented in a future version of the algorithm.
                
                You can ignore covalency by setting check_covalency = False. Don't
                be surprised if the results are weird though!
                """
            )

    def check_structure_for_covalency(self, closest_neighbors):
        """
        This function is designed to check for covalency along the bonds from each
        atom to its nearest neighbors. The NN are defined by Pymatgen's CrystalNN
        function and covalency is described as a maximum in the ELF that is closer
        to the center of the bond than any minimum.

        Args:
            closest_neighbors (dict): A dictionary relating sites to pymatgen
                Neigh objects determined from CrystalNN

        Returns:
            Nothing
        """
        grid = self.grid.copy()
        structure = grid.structure
        for site_index, neighs in closest_neighbors.items():
            # get voxel position from fractional site
            site_voxel_coord = grid.get_voxel_coords_from_index(site_index)
            # iterate over each neighbor bond
            for neigh_index, neigh in enumerate(neighs):
                # Check that we are not looking between two of the same atom. This
                # is likely to have some amount of ELF between the two atoms even
                # if they are not likely to actually be bonded with one another. I
                # don't think this should cause any issues with the algorithm, but
                # !!! it's worth testing
                site_species = structure[site_index].species_string
                neigh_species = structure[neigh["site_index"]].species_string

                if site_species == neigh_species:
                    continue

                # neigh_site_index = neigh["site_index"]
                # if site_index == neigh_site_index:
                #     continue

                neigh_voxel_coord = grid.get_voxel_coords_from_neigh_CrystalNN(neigh)
                values = self.get_partitioning_line_from_voxels(
                    site_voxel_coord, neigh_voxel_coord
                )[1]
                # smooth line
                # smoothed_values = savgol_filter(values,20,3)

                # Now we check for any strong covalency in the bond as in the current version
                # of BadELF, this will break the partitioning scheme
                if self.check_bond_for_covalency(values) is True:
                    # report which atoms the bond was found between

                    # import matplotlib.pyplot as plt
                    # plt.plot(smoothed_values)
                    raise Exception(
                        f"""
                        A maximum in the ELF line between atoms was found closer to the
                        center of the bond than any local minimum.
                        This typically indicates that there is some covalent behavior in
                        your system. Unfortunately, the current version of BadELF does not
                        have a way to partition peaks in the ELF from covalency, though
                        this will hopefully be implemented in a future version of the
                        algorithm.
                        
                        An alternative issue is that you are using a pseudopotential that
                        does not include more than the minimum valence electrons. This will
                        result in ELF values close to 0 around the core of the atom. Make
                        sure you are using suitable pseudopotentials for all of your atoms.
                        We are aware of at least two atoms, Al and B, that do not have a
                        suitable pseudopotential with core electrons in VASP 5.X.X.
                        
                        The bond was found between {site_species} and {neigh_species}.
                        You can ignore covalency by setting check_covalency = False. Don't
                        be surprised if the results are weird though!
                        """
                    )

    def get_site_neighbor_frac(
        self,
        site_cart_coords: ArrayLike,
        neigh_cart_coords: ArrayLike,
        site_symbol: str,
        neigh_symbol: str,
    ):
        """
        Function for getting the fraction of a line betwaeen two sites where
        the ELF is at a minimum.

        Args:
            site_cart_coords (Array | list):
                The cartesian coordinates of the first site.
            neigh_cart_coords (Array | list):
                The cartesian coordinates of the neighboring site.

        returns:
            The minimum point in the ELF between the two sites.

        """
        grid = self.grid.copy()

        site_voxel_coord = grid.get_voxel_coords_from_cart(site_cart_coords)
        neigh_voxel_coord = grid.get_voxel_coords_from_cart(neigh_cart_coords)

        # we need a straight line between these two points.  get list of all ELF values
        elf_coordinates, elf_values = self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord, method="linear"
        )

        # find the minimum position and value along the elf_line
        # the third element is the fractional position, measured from site_voxel_coord
        elf_min_index, elf_min_value, elf_min_frac = self.get_line_minimum_as_frac(
            elf_coordinates,
            elf_values,
            site_symbol,
            neigh_symbol,
        )
        return elf_min_frac

    @staticmethod
    def get_plane_equation(
        plane_point: ArrayLike | list,
        plane_vector: ArrayLike | list,
    ):
        """
        Gets the equation of a plane from a vector normal to the plane and a
        point on the plane

        Args:
            plane_point (ArrayLike): A point on the plane
            plane_vector (ArrayLike): A vector perpendicular to the plane

        Returns:
            The four plane equation coefficients
        """
        x0, y0, z0 = plane_point
        a, b, c = plane_vector
        d = a * x0 + b * y0 + c * z0
        return a, b, c, d

    @classmethod
    def find_intersection_point(
        cls,
        plane1: ArrayLike | list,
        plane2: ArrayLike | list,
        plane3: ArrayLike | list,
    ):
        """
        Finds the point at which three planes intersect if it exists

        Args:
            plane1 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.
            plane2 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.
            plane3 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.

        Returns:
            The point at which the planes intersect as an array
        """

        a1, b1, c1, d1 = cls.get_plane_equation(plane1[:3], plane1[3:])
        a2, b2, c2, d2 = cls.get_plane_equation(plane2[:3], plane2[3:])
        a3, b3, c3, d3 = cls.get_plane_equation(plane3[:3], plane3[3:])

        A = np.array([[a1, b1, c1], [a2, b2, c2], [a3, b3, c3]])
        b = np.array([d1, d2, d3])

        # Solve the system of equations
        intersection_point = np.linalg.solve(A, b)

        return intersection_point

    @classmethod
    def get_important_planes(cls, planes: list):
        """
        Gets a list of planes that make up a 3D polyhedral.

        Args:
            planes (list): The list of planes to check

        Returns:
            A list of plane intercepts and a list of important planes as arrays
            with the first 3 values being the vector and the second three values
            being the point.
        """
        # create list for points where planes intersect and for important planes
        intercepts = []
        important_planes = []

        # iterate through each set of 3 planes
        for combination in combinations(planes, 3):
            # try to find an intersection point. We do a try except because if two
            # of the planes are parallel there wont be an intersect point
            try:
                intercept = cls.find_intersection_point(
                    combination[0], combination[1], combination[2]
                )
            except:
                continue

            # Check if the points are one or within the convex shape defined by the
            # planes. Assume this is true at first
            important_intercept = True
            # Check each plane versus the intercept point. If we plug the point into
            # the plane equation it should return as 0 or positive if it is within the
            # shape?
            for plane in planes:
                sign, dist = cls.get_plane_sign(
                    point=intercept, plane_vector=plane[3:], plane_point=plane[:3]
                )
                if sign in ["positive", "zero"]:
                    pass
                else:
                    important_intercept = False
                    break
            # If the point is bound by all planes, it is an important intercept.
            # append it to our list. Also append any new important planes
            if important_intercept:
                intercepts.append(intercept)
                for plane in combination:
                    point = plane[:3]
                    vector = plane[3:]
                    repeat_plane = False
                    # check if plane already exists in list
                    for plane1 in important_planes:
                        point1 = plane1[:3]
                        vector1 = plane1[3:]

                        # Check if these planes have the same point and vector. If
                        # they do, indicate that this is a repeate plane
                        if np.array_equal(point, point1) and np.array_equal(
                            vector, vector1
                        ):
                            repeat_plane = True
                            break
                    # If this isn't a repeat plane, add it to our important planes list
                    if not repeat_plane:
                        important_planes.append(plane)

        return intercepts, important_planes

    def get_partitioning(self, check_for_covalency: bool = True):
        """
        Gets the partitioning planes for each atom as well as some other useful
        information.

        Args:
            check_for_covalency (bool):
                Whether to check the structure for signs of covalency. This can
                be turned off, but it may give strange results!

        Returns:
            A dictionary where the keys are site indices and the values
            are neighbor dictionaries containing information on the partitioning
            planes.
        """
        logging.info("Beginning partitioning")
        logging.info("Checking structure for covalency")
        if check_for_covalency:
            closest_neighbors = self.get_closest_neighbors()
            self.check_structure_for_covalency(closest_neighbors)
            self.check_closest_neighbor_for_same_type(closest_neighbors)

        grid = self.grid
        structure = grid.structure

        logging.info("Getting neighboring atoms for each site in structure")
        # the 50 nearest neighbors for each atom in the structure
        nearest_neighbors = self.get_set_number_of_neighbors(50)

        # Create a dataframe to store the important information about each site/neighbor pair
        site_neigh_pairs = pd.DataFrame(
            columns=[
                "site_index",
                "neigh_index",
                "neigh",
                "site_symbol",
                "neigh_symbol",
                "site_coords",
                "neigh_coords",
                "dist",
            ]
        )
        # Add the important information for each site/neighbor pair
        for site, neighs in enumerate(nearest_neighbors):
            site_cart_coords = structure[site].coords
            for neigh in neighs:
                neigh_cart_coords = neigh.coords
                dist = round(math.dist(site_cart_coords, neigh_cart_coords), 5)
                row = [
                    site,
                    neigh.index,
                    neigh,
                    structure[site].species_string,
                    neigh.species_string,
                    site_cart_coords,
                    neigh_cart_coords,
                    dist,
                ]
                site_neigh_pairs.loc[len(site_neigh_pairs)] = row

        site_neigh_pairs["partitioning_frac"] = None
        site_neigh_pairs["radius"] = None
        # create a new dataframe containing only unique pairs
        unique_pairs = site_neigh_pairs.drop_duplicates(
            subset=["site_symbol", "neigh_symbol", "dist"]
        )

        logging.info("Finding partitioning planes")
        # Get partitioning frac for each unique site_neighbor pair
        for index, row in tqdm(unique_pairs.iterrows()):
            # Check if we've already found the frac for this row
            if row["partitioning_frac"] is None:
                # get coords of each site and its neighbor
                site_cart_coords = row["site_coords"]
                # site_voxel_coords = grid.get_voxel_coords_from_cart(site_cart_coords)
                neigh_cart_coords = row["neigh_coords"]
                # neigh_voxel_coords = grid.get_voxel_coords_from_cart(neigh_cart_coords)

                site_symbol = row["site_symbol"]
                neigh_symbol = row["neigh_symbol"]
                dist = row["dist"]

                # get fraction along line where the min is located
                frac = self.get_site_neighbor_frac(
                    site_cart_coords, neigh_cart_coords, site_symbol, neigh_symbol
                )
                radius = frac * dist
                reverse_frac = 1 - frac
                reverse_radius = reverse_frac * dist

                # create search to find rows with same symbol set and reverse symbol set.
                reverse_condition = (
                    (unique_pairs["site_symbol"] == neigh_symbol)
                    & (unique_pairs["neigh_symbol"] == site_symbol)
                    & (unique_pairs["dist"] == dist)
                )
                # assign the fraction along the line and distance to each unique
                # site neighbor pair. We do this in the loop so that the reverse
                # assignments don't need to be repeated
                unique_pairs.at[index, "partitioning_frac"] = frac
                unique_pairs.loc[reverse_condition, "partitioning_frac"] = reverse_frac

                # create another search condition for the full dataframe of site-neighbor pairs
                search_condition1 = (
                    (site_neigh_pairs["site_symbol"] == site_symbol)
                    & (site_neigh_pairs["neigh_symbol"] == neigh_symbol)
                    & (site_neigh_pairs["dist"] == dist)
                )
                reverse_condition1 = (
                    (site_neigh_pairs["site_symbol"] == neigh_symbol)
                    & (site_neigh_pairs["neigh_symbol"] == site_symbol)
                    & (site_neigh_pairs["dist"] == dist)
                )

                # Assign the fraction along the line and distance to ever
                # site neighbor pair.
                site_neigh_pairs.loc[
                    search_condition1, ["partitioning_frac", "radius"]
                ] = (frac, radius)
                site_neigh_pairs.loc[
                    reverse_condition1, ["partitioning_frac", "radius"]
                ] = (reverse_frac, reverse_radius)

        # Get the site and neighbor coords in arrays to make calculations easier
        site_coords = np.array(site_neigh_pairs["site_coords"].to_list())
        neigh_coords = np.array(site_neigh_pairs["neigh_coords"].to_list())
        fracs = site_neigh_pairs["partitioning_frac"].to_numpy()

        # Calculate the plane points and vectors
        vectors = neigh_coords - site_coords
        magnitudes = np.linalg.norm(vectors, axis=1)
        unit_vectors = vectors / magnitudes[:, np.newaxis]
        plane_points = vectors * fracs[:, np.newaxis] + site_coords

        # Add plane points and vectors to full dataframe
        site_neigh_pairs["plane_points"] = list(plane_points)
        site_neigh_pairs["plane_vectors"] = list(unit_vectors)
        # Combine plane points and vectors into one array
        planes = np.concatenate((plane_points, unit_vectors), axis=1)
        all_atom_planes = np.split(planes, len(structure))

        logging.info("Reducing to necessary partitioning planes")
        partitioning_planes = []
        for atom_planes in all_atom_planes:
            important_planes = self.get_important_planes(atom_planes)[1]
            partitioning_planes.append(important_planes)

        # Create a dict to store the final dataframe for each atom
        partitioning = {}
        for site_index, important_planes in tqdm(enumerate(partitioning_planes)):
            partitioning_df = site_neigh_pairs.loc[
                site_neigh_pairs["site_index"] == site_index
            ]
            important_indices = []
            for index, row in partitioning_df.iterrows():
                # get the associated point and vector for the partitioning plane in
                # this neighbor.
                plane_point = np.array(row["plane_points"])
                plane_vector = np.array(row["plane_vectors"])
                # Check if this plane exists in our list of important planes. If it
                # does than we'll add this row to our important partitioning planes
                # list
                for plane in important_planes:
                    point = plane[:3]
                    vector = plane[3:]
                    if np.array_equal(plane_point, point) and np.array_equal(
                        plane_vector, vector
                    ):
                        important_indices.append(index)
            partitioning_df = partitioning_df[
                partitioning_df.index.isin(important_indices)
            ]
            partitioning[site_index] = partitioning_df
        logging.info("Finished with partitioning")
        return partitioning

    def plot_partitioning_results(
        self,
        partition_results: dict = None,
    ):
        """
        Plots the unit cell and partitioning planes from the partitioning
        results.

        Args:
            partition_results (dict): The results from running the partitioning
                algorithm
        """
        if partition_results is None:
            partition_results = self.get_partitioning()

        # Create a matplotlib plot
        import matplotlib
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")

        # Get the vertices of the polygons formed by the partitioning around
        # each atom
        atoms_polygon_nodes = {}
        for site_index, neighbor_df in partition_results.items():
            planes = []
            for neigh_index, row in neighbor_df.iterrows():
                plane_point = list(row["plane_points"])
                plane_vector = list(row["plane_vectors"])
                plane = plane_vector + plane_point
                planes.append(plane)
            intercepts, _ = self.get_important_planes(planes)
            atoms_polygon_nodes[site_index] = intercepts

        # get the structure and species
        structure = self.grid.structure
        species = structure.symbol_set

        # plot the unit cell
        self.grid._plot_unit_cell(ax, fig)

        # get a color map to match all same atoms to the same color
        color_map = matplotlib.colormaps.get_cmap("tab10")

        # go through each atom species, set the unique color
        for i, specie in enumerate(species):
            color = color_map(i)
            site_indices = structure.indices_from_symbol(specie)
            # for each site, loop through and plot the surfaces.
            for site in site_indices:
                hull = ConvexHull(atoms_polygon_nodes[site])
                triangles = hull.simplices

                x = []
                y = []
                z = []
                for point in atoms_polygon_nodes[site]:
                    x.append(point[0])
                    y.append(point[1])
                    z.append(point[2])

                ax.plot_trisurf(x, y, z, triangles=triangles, color=color)



# -*- coding: utf-8 -*-

import numpy as np
from scipy.spatial.distance import cosine


class ASEFingerprint:
    # ASE cites this as a rewrite of the USPEX fingerprint
    # https://gitlab.com/ase/ase/-/blob/master/ase/ga/ofp_comparator.py

    def __init__(self, composition, on_fail, max_fails, initial_structures=[]):
        self.on_fail = on_fail

        # we need the cumulative max_fails list
        # for example, the max_fails=[20,40,10] needs to be converted to [20,60,70]
        # if infinite attempts are allowed, be sure to use float('inf') as the value
        # This list is more efficient to work with in the point_next_step()
        #!!! should I make this numpy for speed below?
        self.max_fails_cum = [sum(max_fails[: x + 1]) for x in range(len(max_fails))]

        # we can assume the user has ASE installed because it is a dependency of PyMatgen
        #!!! it looks like the ase.ga module is actively changing so version may introduce errors
        from ase.ga.ofp_comparator import OFPComparator

        self.featurizer = OFPComparator(  # n_top=None, # atoms to optimize. None is all of them
            # dE=1.0, # energy difference to be considered different (not used here b/c we dont have a app)
            # cos_dist_max=5e-3, # cosine distance for structure to be considered different
            # rcut=20., # cutoff radius for the fingerprint
            # binwidth=0.05, # bin width (in Angstrom) over which fingerprints are discretized
            # sigma=0.02, # smearing of the guassian in fingerprint
            # nsigma=4, # distance where smearing is cut off
            # pbc=True, # whether to apply periodic boundry conditions
            # maxdims=None, # not used if we have a 3d crystal
            # recalculate=False # wether to recalculate fingerprints in ASE database
        )

        # we also need to convert pymatgen Structures to ase Atoms below
        #!!! is it faster and more memory efficient to import below?
        from pymatgen.io.ase import AseAtomsAdaptor

        self.adaptor = AseAtomsAdaptor

        # generate the fingerprint for each of the initial input structures
        self.fingerprint_database = []
        for structure in initial_structures:
            # convert the structure to ase Atoms object
            structure_ase = self.adaptor.get_atoms(structure)

            # make the fingerprint
            # ase uses .looks_like() to compare two structures, but we want the lower-level function ._take_fingerprints()
            output = self.featurizer._take_fingerprints(structure_ase)
            # the output is a tuple of...
            # (1) dictionary with keys of each element combo mapped to the fingerprint
            # (2) dictionary with keys of element to the indicies of that element in the Atoms object
            # so I want to make the fingerprint a 1D array so I go through the output and merge these.
            # make a list of each element-element fingerprint
            fingerprints = [output[0][key] for key in output[0]]
            # now combine the fingerprints into a single 1D array
            fingerprint = np.hstack(fingerprints)

            # add to the database
            self.fingerprint_database.append(fingerprint)
        # now make the datebase into a numpy array for speed
        self.fingerprint_database = np.array(self.fingerprint_database)

    def check_structure(self, structure, tolerance=5e-3):
        # convert the structure to ase Atoms object
        structure_ase = self.adaptor.get_atoms(structure)

        # make the fingerprint
        # ase uses .looks_like() to compare two structures, but we want the lower-level function ._take_fingerprints()
        output = self.featurizer._take_fingerprints(structure_ase)
        # the output is a tuple of...
        # (1) dictionary with keys of each element combo mapped to the fingerprint
        # (2) dictionary with keys of element to the indicies of that element in the Atoms object
        # so I want to make the fingerprint a 1D array so I go through the output and merge these.
        # make a list of each element-element fingerprint
        fingerprints = [output[0][key] for key in output[0]]
        # now combine the fingerprints into a single 1D array
        fingerprint1 = np.hstack(fingerprints)

        # we now want to get the distance of this fingerprint relative to all others
        # if the distance is within the specified tolerance, then the structures are too similar - we return false
        # if we make it through all structures and no distance is below the tolerance, we are good to go!
        for fingerprint2 in self.fingerprint_database:
            # we want the cosine distance
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cosine.html
            # I use the numpy version of this for speed
            distance = cosine(fingerprint1, fingerprint2)
            if distance < tolerance:
                # we can end the whole function as soon as one structure is deemed too similar
                return False
            else:
                continue

        # if we finished the loop above, we can add this structure to the database and return a success
        # this method I use to append is pretty slow for numpy. Is there a better way to do this?
        if self.fingerprint_database.size == 0:
            self.fingerprint_database = np.array([fingerprint1])
        else:
            self.fingerprint_database = np.append(
                self.fingerprint_database, [fingerprint1], axis=0
            )

        return True

    def point_next_step(self, attempt=0):
        # based on the attempt number and the max_fails list, see what stage of on_fail we are on
        for i, attempt_cutoff in enumerate(self.max_fails_cum):
            # if we are below the attempt cutoff, return the corresponding on_fail object
            if attempt <= attempt_cutoff:
                return self.on_fail[i]

        # If we make it through all of the cutoff list, that means we exceeded the maximum attempts allowed
        # In this case, we don't return return an object from the on_fail list
        return False



# -*- coding: utf-8 -*-

import requests

from simmate.engine import Workflow


class WebApiWorkflow(Workflow):
    """
    The base Web API workflow that many other workflows inherit from. This class
    encapulates logic for calling external APIs (e.g. a REST API) and handling
    common functionality and errors (e.g. chunking the cal, connection errors, etc.).

    Most commonly, this class is used for Deployed AI/ML models that are hosted
    as microservices elsewhere.
    """

    use_database = False
    _parameter_methods = Workflow._parameter_methods + ["_get_payload"]

    api_url: str = None
    """
    The full REST API endpoint that should be called for this method
    """

    repo_url: str = None
    """
    Because Web APIs are often hosted as a microservice elsewhere, their code
    is hosted elsewhere as well. This points to the Web API's source code.
    """

    maintainer: str = None
    """
    The current maintainer of this API endpoint and service.
    """

    chunk_input: str = None  # e.g. "molecules"
    chunk_size: int = None  # e.g. 10000

    timeout: float = None  # per chunk

    @classmethod
    def run_config(cls, **kwargs):
        if not cls.api_url:
            raise Exception(
                "'DeployedModel' workflows require the 'api_url' attribute to be set"
            )

        # TODO: add in chunking functionality for call api's with massive
        # datasets (e.g. >1mil molecules)

        response = requests.post(
            cls.api_url,
            json=cls._get_payload(**kwargs),
            timeout=cls.timeout,
        )
        response_data = response.json()
        return cls._reformat_response(response_data)

    @staticmethod
    def _get_payload(**kwargs) -> dict:
        # by default, we just pass all input parameters as dictionary. In the
        # large majority of cases, this method should be modified though.
        return kwargs

    @staticmethod
    def _reformat_response(response: dict) -> dict:
        # by default, we don't modify the returned data. But if the model returns
        # data in an undesirable
        return response



# -*- coding: utf-8 -*-

import logging
import traceback
import warnings
from pathlib import Path

from simmate.database.base_data_types import Calculation, table_column
from simmate.toolkit import Composition
from simmate.utilities import get_chemical_subsystems, get_directory

# BUG: This prints a tqdm error so we silence it here.
with warnings.catch_warnings(record=True):
    from pymatgen.analysis.phase_diagram import PhaseDiagram


class ChemicalSystemSearch(Calculation):
    class Meta:
        app_label = "workflows"

    chemical_system = table_column.CharField(max_length=10, null=True, blank=True)
    subworkflow_name = table_column.CharField(max_length=200, null=True, blank=True)
    subworkflow_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    max_atoms = table_column.IntegerField(null=True, blank=True)
    max_stoich_factor = table_column.IntegerField(null=True, blank=True)
    singleshot_sources = table_column.JSONField(default=list, null=True, blank=True)

    # DEV NOTE: many of the methods below are copy/pasted from the fixed
    # composition table and functionality should be merged in the future.

    # -------------------------------------------------------------------------
    # Core methods that help grab key information about the search
    # -------------------------------------------------------------------------

    def to_toolkit(self) -> PhaseDiagram:
        phase_diagram = self.individuals_datatable.get_phase_diagram(
            chemical_system=self.chemical_system,
            workflow_name=self.subworkflow.name_full,
        )
        return phase_diagram

    @property
    def chemical_system_cleaned(self):
        # simply ordered elements in alphabetical order as that is how they
        # are stored in the database
        endpoint_compositions = [
            Composition(sys) for sys in self.chemical_system.split("-")
        ]
        elements = []
        for comp in endpoint_compositions:
            for element in comp.elements:
                element = str(element)
                if element not in elements:
                    elements.append(element)
        elements.sort()
        chemical_system = "-".join(elements)
        return chemical_system

    @property
    def chemical_subsystems(self):
        return get_chemical_subsystems(self.chemical_system_cleaned)

    @property
    def subworkflow(self):
        from simmate.workflows.utilities import get_workflow

        if self.subworkflow_name == "relaxation.vasp.staged":
            return get_workflow(self.subworkflow_name)
        else:
            raise Exception(
                "Only `relaxation.vasp.staged` is supported in early testing"
            )

    @property
    def individuals_datatable(self):
        return self.subworkflow.database_table

    @property
    def individuals(self):
        return self.individuals_datatable.objects.filter(
            chemical_system__in=self.chemical_subsystems,
            workflow_name=self.subworkflow_name,
        )

    @property
    def individuals_completed(self):
        return self.individuals.filter(energy_per_atom__isnull=False)

    @property
    def individuals_incomplete(self):
        # If there is an energy_per_atom, we can treat the calculation as completed
        return self.individuals.filter(energy_per_atom__isnull=True)

    @property
    def stable_structures(self):
        structures = self.individuals_completed.filter(energy_above_hull=0).to_toolkit()
        return structures

    @property
    def best_structure_for_each_composition(self):
        # BUG: for sqlite, you can't use distinct.
        # structures = (
        #     self.individuals_completed.order_by("energy_above_hull")
        #     .distinct("formula_reduced")
        #     .to_toolkit()
        # )

        # Instead, I just use pymatgen. This is slower but still works
        phase_diagram = self.to_toolkit()

        structures = [
            self.individuals.get(id=int(e.entry_id.split("=")[-1]))
            for e in phase_diagram.qhull_entries
        ]
        return [s.to_toolkit() for s in structures]

    def update_stabilities(self):
        self.individuals_datatable.update_chemical_system_stabilities(
            chemical_system=self.chemical_system_cleaned,
            workflow_name=self.subworkflow.name_full,
        )

    # -------------------------------------------------------------------------
    # Writing CSVs summaries and CIFs of best structures
    # -------------------------------------------------------------------------

    def write_output_summary(self, directory):
        # If the output fails to write, we have a non-critical issue that
        # doesn't affect the search. We therefore don't want to raise an
        # error here -- but instead warn the user and then continue the search
        try:
            if not self.individuals_completed.exists():
                logging.info("No structures completed yet. Skipping output writing.")
                return

            logging.info(f"Writing search summary to {directory}")

            super().write_output_summary(directory=directory)

            # update all chemical stabilites before creating the output files
            self.update_stabilities()

            self.individuals_datatable.write_hull_diagram_plot(
                chemical_system=self.chemical_system_cleaned,
                workflow_name=self.subworkflow.name_full,
                directory=directory,
                show_unstable_up_to=5,
            )

            stable_dir = get_directory(directory / "stable_structures")
            self.write_stable_structures(stable_dir)

            all_comps_dir = get_directory(directory / "best_structure_per_composition")
            self.write_stable_structures(all_comps_dir, include_best_metastable=True)

        except Exception as error:
            if (
                isinstance(error, ValueError)
                and "no entries for the terminal elements" in error.args[0]
            ):
                logging.warning(
                    "The convex hull and structure stabilities cannot be calculated "
                    "without terminal elements. Either manually submit pure-element "
                    "structures to your subworkflow, or make sure you run your "
                    "search with singleshot sources active (the default) AND "
                    "your database populated with third-party data. Output files "
                    "will not be written until this is done."
                )

            else:
                logging.warning(
                    "Failed to write the output summary. This issue will be silenced "
                    "to avoid stopping the search. But please report the following "
                    "error to our github: https://github.com/jacksund/simmate/issues/"
                )

                # prints the most recent exception traceback
                traceback.print_exc()

    def write_stable_structures(
        self,
        directory: Path,
        include_best_metastable: bool = False,
    ):
        # if the directory is filled, we need to delete all the files
        # before writing the new ones.
        for file in directory.iterdir():
            try:
                file.unlink()
            except OSError:
                logging.warning("Unable to delete a CIF file: {file}")
                logging.warning(
                    "Updating the 'best structures' directory involves deleting "
                    "and re-writing all CIF files each cycle. If you have a file "
                    "open while this step occurs, then you'll see this warning."
                    "Close your file for this to go away."
                )

        structures = (
            self.stable_structures
            if not include_best_metastable
            else self.best_structure_for_each_composition
        )

        for rank, structure in enumerate(structures):
            rank_cleaned = str(rank).zfill(2)  # converts 1 to 01
            structure_filename = directory / (
                f"rank-{str(rank_cleaned)}__"
                + f"id-{structure.database_object.id}"
                + f"id-{structure.composition.reduced_formula}.cif"
            )
            structure.to(filename=str(structure_filename), fmt="cif")

    def write_individuals_completed_full(self, directory: Path):
        columns = self.individuals_datatable.get_column_names()
        columns.remove("structure")
        df = self.individuals_completed.defer("structure").to_dataframe(columns)
        csv_filename = directory / "individuals_completed__ALLDATA.csv"
        df.to_csv(csv_filename)

    def write_individuals_completed(self, directory: Path):
        columns = [
            "id",
            "energy_per_atom",
            "finished_at",
            "source",
            "spacegroup__number",
        ]
        df = (
            self.individuals_completed.order_by(self.fitness_field)
            .only(*columns)
            .to_dataframe(columns)
        )
        # label the index column
        df.index.name = "rank"

        # make the timestamps easier to read
        def format_date(date):
            return date.strftime("%Y-%m-%d %H:%M:%S")

        df["finished_at"] = df.finished_at.apply(format_date)

        def format_parents(source):
            return source.get("parent_ids", None) if source else None

        df["parent_ids"] = df.source.apply(format_parents)

        def format_source(source):
            return (
                None
                if not source
                else source.get("creator", None) or source.get("transformation", None)
            )

        df["source"] = df.source.apply(format_source)

        # shorten the column name for easier reading
        df.rename(columns={"spacegroup__number": "spacegroup"}, inplace=True)

        md_filename = directory / "individuals_completed.md"
        df.to_markdown(md_filename)



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)
from simmate.apps.vasp.workflows.electronic_structure import VaspDensityOfStates


class ElectronicStructure__Vasp__MatprojDensityOfStates(
    VaspDensityOfStates, Relaxation__Vasp__Matproj
):
    """
    This task is a reimplementation of pymatgen's
    [MPNonSCFSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPNonSCFSet)
    with mode="uniform".

    Calculates the band structure using Materials Project HSE settings.
    """

    parent_workflows = ["electronic-structure.vasp.matproj-full"]
    _incar_updates = dict(
        IBRION=-1,
        LCHARG=False,
        LORBIT=11,
        LWAVE=False,
        NSW=0,
        ISYM=2,
        ICHARG=11,
        ISMEAR=-5,
        NEDOS=2001,
        MAGMOM__smart_magmom="__remove__",
    )



# -*- coding: utf-8 -*-

import logging

from simmate.apps.quantum_espresso.inputs.potentials_sssp import check_psuedo_setup
from simmate.configuration import settings
from simmate.configuration.utilities import (
    check_app_reg,
    check_command_exists,
    show_test_results,
)


def test_config(run_calcs: bool = False):
    """
    Checks to see if Quantum Espresso is configured correctly
    """

    # 1 - check that Simmate app is registered
    is_registered = check_app_reg("simmate.apps.configs.QuantumEspressoConfig")

    # 2 - check for command
    use_docker = settings.quantum_espresso.docker.enable
    command = "pw.x" if not use_docker else "docker"
    found_command = check_command_exists(command)

    # 3 - check for psuedopotentials
    has_puedos = check_psuedo_setup()
    if has_puedos:
        logging.info("Psuedopotentials (SSSP) found :heavy_check_mark:")
    else:
        logging.warning("Psuedopotentials (SSSP) not found.")

    # 4 - run some sample workflows
    if run_calcs:
        raise NotImplementedError("This test has not been added yet.")

    # 5 - read out result of all tests
    passed_all = bool(is_registered and found_command and has_puedos)
    show_test_results("Quantum Espresso", passed_all)



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .mit import (
    Diffusion__Vasp__NebAllPathsMit,
    Diffusion__Vasp__NebFromEndpointsMit,
    Diffusion__Vasp__NebFromImagesMit,
    Diffusion__Vasp__NebFromImagesMvlCi,
    Diffusion__Vasp__NebSinglePathMit,
)



# -*- coding: utf-8 -*-

import logging

import pytest

from simmate.workflows.utilities import get_all_workflow_names, get_workflow


# @pytest.mark.prefect_db
@pytest.mark.vasp
@pytest.mark.django_db
def test_all_workflow_runs(tmp_path, sample_structures):
    # For testing, look at the NaCl rocksalt primitive structure
    structure = sample_structures["NaCl_mp-22862_primitive"]

    # -------------
    # https://stackoverflow.com/questions/41742317/
    import contextlib
    import os
    from pathlib import Path

    @contextlib.contextmanager
    def working_directory(path):
        """Changes working directory and returns to previous on exit."""
        prev_cwd = Path.cwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(prev_cwd)

    # -----------------

    with working_directory(tmp_path):
        successful_flows = []

        # TEST CUSTOMIZED FLOW
        workflow_name = "customized.toolkit.user-config"
        workflow = get_workflow(workflow_name)
        state = workflow.run(
            workflow_base="relaxation.vasp.quality00",
            input_parameters={
                "structure": structure,
                "command": "mpirun -n 14 vasp_std > vasp.out",
            },
            updated_settings={
                "incar": {"NPAR": 1, "ENCUT": 600},
            },
        )
        if state.is_completed():
            successful_flows.append(workflow_name)

        # TEST S3Workflows
        s3_flows = [
            "relaxation.vasp.matproj",
            "relaxation.vasp.mit",
            "relaxation.vasp.mvl-neb-endpoint",
            "relaxation.vasp.quality00",
            "relaxation.vasp.quality01",
            "relaxation.vasp.quality02",
            "relaxation.vasp.quality03",
            "relaxation.vasp.quality04",
            "relaxation.vasp.staged",
            "relaxation.vasp.matproj-metal",
            "relaxation.vasp.matproj-scan",  # slow
            "relaxation.vasp.matproj-hse",  # slow
            "relaxation.vasp.mvl-grainboundary",
            "relaxation.vasp.mvl-slab",
            "relaxation.vasp.warren-lab-hse",
            "relaxation.vasp.warren-lab-hse-with-wavecar",
            "relaxation.vasp.warren-lab-hsesol",
            "relaxation.vasp.warren-lab-pbe",
            "relaxation.vasp.warren-lab-pbe-metal",
            "relaxation.vasp.warren-lab-pbe-with-wavecar",
            "relaxation.vasp.warren-lab-pbesol",
            "relaxation.vasp.warren-lab-scan",
            "static-energy.vasp.matproj",
            "static-energy.vasp.mit",
            "static-energy.vasp.mvl-neb-endpoint",
            "static-energy.vasp.quality04",
            "static-energy.vasp.matproj-hse",  # slow
            "static-energy.vasp.matproj-scan",  # slow
            "static-energy.vasp.warren-lab-hse",
            "static-energy.vasp.warren-lab-hsesol",
            "static-energy.vasp.warren-lab-pbe",
            "static-energy.vasp.warren-lab-pbe-metal",
            "static-energy.vasp.warren-lab-pbesol",
            "static-energy.vasp.warren-lab-prebadelf-hse",
            "static-energy.vasp.warren-lab-prebadelf-pbesol",
            "static-energy.vasp.warren-lab-scan",
            "population-analysis.vasp-bader.bader-matproj",
            "population-analysis.vasp.elf-matproj",
            "electronic-structure.vasp.matproj-full",
            "electronic-structure.vasp.matproj-hse-full",  # slow
        ]
        for workflow_name in s3_flows:
            workflow = get_workflow(workflow_name)
            state = workflow.run(
                structure=structure,
                command="mpirun -n 14 vasp_std > vasp.out",
            )
            if state.is_completed():
                successful_flows.append(workflow_name)

        # TEST MD FLOW
        dynamics_flows = [
            "dynamics.vasp.mit",
            "dynamics.vasp.matproj",
            "dynamics.vasp.mvl-npt",
        ]
        for workflow_name in dynamics_flows:
            workflow = get_workflow(workflow_name)
            state = workflow.run(
                structure=structure,
                command="mpirun -n 14 vasp_std > vasp.out",
                nsteps=50,
                temperature_start=400,
                temperature_end=400,
            )
            if state.is_completed():
                successful_flows.append(workflow_name)

        # TEST NEB FLOWS
        # For diffusion, we look at I- diffusion in Y2CI2 (takes roughly 1 hr)
        structure = sample_structures["Y2CI2_mp-1206803_primitive"]
        workflow_name = "diffusion.vasp.neb-all-paths-mit"
        workflow = get_workflow(workflow_name)
        state = workflow.run(
            structure=structure,
            migrating_specie="I",
            command="mpirun -n 14 vasp_std > vasp.out",
            nimages=1,
            min_atoms=10,
            max_atoms=25,
            min_length=4,
        )
        state.result()
        if state.is_completed():
            successful_flows.append(workflow_name)

        # TEST BadELF FLOWS
        structure = sample_structures["Ca2N_mp-2686_primitive"]
        workflow_name = "bad-elf-analysis.badelf.badelf-test"
        workflow = get_workflow(workflow_name)

        # This workflow runs a static energy and then badelf calculation. The
        # hse version is identical, but uses a different static energy step
        # which is tested above. However, there are 3 different algorithms
        # which can be used for partitioning, each of which needs to be tested
        algorithms = ["badelf", "voronelf", "zero-flux"]
        for algorithm in algorithms:
            state = workflow.run(
                structure=structure,
                command="mpirun -n 14 vasp_std > vasp.out",
                algorithm=algorithm,
                cores=14,
                check_for_covalency=False,
            )
            state.result()
            if state.is_completed():
                successful_flows.append(workflow_name)

    # check which flows either (1) failed or (2) weren't tested
    all_flows = get_all_workflow_names()
    missing_failed_flows = list(set(all_flows) - set(successful_flows))
    missing_failed_flows.sort()

    logging.info(
        f"The following workflows either failed or were not tested: {missing_failed_flows}"
    )
    # assert missing_failed_flows == []



# Generated by Django 4.2.2 on 2023-07-21 16:22

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="WorkItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("tags", models.JSONField(default=list)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("P", "Pending"),
                            ("R", "Running"),
                            ("C", "Cancelled"),
                            ("E", "Errored"),
                            ("F", "Finished"),
                        ],
                        default="P",
                        max_length=1,
                    ),
                ),
                ("command_not_found_failures", models.IntegerField(default=0)),
                ("fxn", models.BinaryField()),
                ("args", models.BinaryField(default=b"\x80\x05]\x94.")),
                ("kwargs", models.BinaryField(default=b"\x80\x05}\x94.")),
                ("result_binary", models.BinaryField(blank=True, null=True)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-

import importlib

from django.contrib.auth.decorators import login_required
from django.core.exceptions import PermissionDenied
from django.db.models import F
from django.shortcuts import render

from simmate.configuration import settings
from simmate.database.third_parties import (
    AflowStructure,
    CodStructure,
    JarvisStructure,
    MatprojStructure,
    OqmdStructure,
)
from simmate.utilities import get_app_submodule, get_class
from simmate.website.data_explorer.forms import ChemicalSystemForm

# -----------------------------------------------------------------------------

# this section is wacky because we want to dynamically set the home and profile
# fxn/views using either a default or some env variable (which allows people
# to override it)


@login_required
def profile_default_view(request):
    # !!! For future reference, you can grab user-associated data via...
    # data = request.user.relateddata.all()

    context = {}
    template = "account/profile.html"
    return render(request, template, context)


@login_required
def profile(request):
    if not settings.website.profile_view:
        return profile_default_view(request)  # This is the function above
    else:
        # we assume the view is named "profile" inside this module
        profile_module = importlib.import_module(settings.website.profile_view)
        profile_view = getattr(profile_module, "profile")
        return profile_view(request)


def home_default_view(request):
    # The default homepage is a bulk query for crystal structures. For internal
    # websites, a different homepage (e.g. a chatbot) is used instead.

    # The home page is also an html "form" because users submit queries from
    # here. So we need to handle form submissions properly.

    # first check if the webpage is accessed via a POST method
    if request.method == "POST":
        # if it is, that means a user is trying to submit a query
        # let's grab the data and validate it before running the query
        form = ChemicalSystemForm(request.POST)
        # see if all of the data is valid
        if form.is_valid():
            # grab the cleaned data from the form
            cleaned_data = form.cleaned_data

            # check which databases the user wants to search and collect the
            # proper models into a list that we query with below
            databases_to_search = []
            for database, database_model in (
                ("aflow", AflowStructure),
                ("cod", CodStructure),
                ("jarvis", JarvisStructure),
                ("materials_project", MatprojStructure),
                ("oqmd", OqmdStructure),
            ):
                # if the user requested this database, the value will be true
                if cleaned_data[database]:
                    databases_to_search.append(database_model)

            # Note, this value is converted to a list of systems once cleaned.
            chemical_systems = cleaned_data["chemical_system"]

            # now go through each database, search for the requested system
            # and then pool them all together. We limit each database to 50
            # results. Note the += below means we compile into a single list
            structures = []
            nstructures_possible = 0
            for database_model in databases_to_search:
                # Now we can make the query! We also dont want to load the
                # structure json -- so that everything runs faster.
                search_results = database_model.objects.filter(
                    chemical_system__in=chemical_systems
                ).defer("structure")

                # if the database provides the hull energy, we want to sort
                # the structures by that (putting highest priority on stable ones)
                if hasattr(database_model, "energy_above_hull"):
                    # if there isn't a hull energy value, place these last
                    search_results = search_results.order_by(
                        F("energy_above_hull").asc(nulls_last=True)
                    )

                # now add the search results to the output
                # for performance, we limit each database to 50 structures
                structures += search_results.all()[:50]

                # We also tell the user how many results are possible if there
                # wasn't any limit on the structures returned
                nstructures_possible += database_model.objects.filter(
                    chemical_system__in=chemical_systems,
                ).count()

    # if the page is grabbed via a 'GET' method, send an empty form
    else:
        # otherwise we are giving an empty form and no result structures
        form = ChemicalSystemForm()
        structures = None
        nstructures_possible = None

    # now let's put the data and template together to send the user
    context = {
        "chemical_system_form": form,
        "structures": structures,
        "nstructures_possible": nstructures_possible,
    }
    template = "home/home.html"
    return render(request, template, context)


def home(request):
    if not settings.website.home_view:
        return home_default_view(request)  # This is the function above
    else:
        # we assume the view is named "home" inside this module
        home_module = importlib.import_module(settings.website.home_view)
        home_view = getattr(home_module, "home")
        return home_view(request)


# -----------------------------------------------------------------------------


def loginstatus(request):
    context = {}
    template = "account/loginstatus.html"
    return render(request, template, context)


def permission_denied(request):
    raise PermissionDenied


def apps(request):
    extra_apps = []
    for app_name in settings.apps:
        urls_path = get_app_submodule(app_name, "urls")
        if urls_path:
            app_config = get_class(app_name)
            extra_apps.append(
                {
                    "verbose_name": app_config.verbose_name,
                    "short_name": app_config.name.split(".")[-1],
                    "description_short": app_config.description_short,
                }
            )
    context = {"extra_apps": extra_apps}
    template = "core_components/apps.html"
    return render(request, template, context)


def extras(request):
    context = {}
    template = "core_components/extras.html"
    return render(request, template, context)


def contact(request):
    context = {}
    template = "core_components/contact.html"
    return render(request, template, context)


def about(request):
    context = {}
    template = "core_components/about.html"
    return render(request, template, context)



# -*- coding: utf-8 -*-

import pytest

from simmate.engine import Workflow
from simmate.website.test_app.models import TestCalculation


class DummyProject__DummyCaclulator__DummyPreset(Workflow):
    """
    Minimal example of a workflow
    """

    use_database = False

    # TODO: test database registration and results
    database_table = TestCalculation
    # @classmethod
    # def _save_to_database(cls, result):
    #     pass  # Skip default which is meant for vasp calcs

    @staticmethod
    def run_config(source=None, structure=None, **kwargs):
        return 1 + 2


# copy to variable for shorthand use
DummyFlow = DummyProject__DummyCaclulator__DummyPreset


@pytest.mark.django_db
def test_workflow(tmp_path):
    # Same exact thing but using higher-level method
    state = DummyFlow.run(directory=tmp_path)
    assert state.is_completed()
    assert state.result() == 3

    # testing naming conventions
    assert DummyFlow.name_full == "dummy-project.dummy-caclulator.dummy-preset"
    assert DummyFlow.name_type == "dummy-project"
    assert DummyFlow.name_app == "dummy-caclulator"
    assert DummyFlow.name_preset == "dummy-preset"

    # testing class properties
    assert DummyFlow.description_doc == DummyFlow.__doc__
    assert DummyFlow.description_doc.strip() == "Minimal example of a workflow"
    assert DummyFlow.parameter_names == [
        "compress_output",
        "directory",
        "run_id",
        "source",
        "structure",
    ]
    assert DummyFlow._parameters_to_register == [
        "directory",
        "run_id",
        "source",
    ]
    DummyFlow.show_parameters()  # a print statment w. nothing else to check

    assert isinstance(DummyFlow.get_config(), dict)

    DummyFlow.show_config()  # a print statment w. nothing else to check

    # test file compression
    new_dir = tmp_path / "new_folder"
    state = DummyFlow.run(directory=new_dir, compress_output=True)
    result = state.result()

    # make sure that a "simmate-task-*.zip" archive was created
    assert new_dir.with_suffix(".zip").exists()

    # make sure that a "simmate-task-*" directory was removed
    assert not new_dir.exists()

    # and delete the archive
    new_dir.with_suffix(".zip").unlink()


def test_serialize_parameters():
    class TestParameter1:
        def to_dict(self):
            return {}

    class TestParameter2:
        """requires cloudpickle to serialize"""

        a = 123

    parameters = dict(
        a=TestParameter1(),
        b=TestParameter2(),
    )
    Workflow._serialize_parameters(**parameters)


def test_deserialize_parameters(mocker):
    # -------
    # We don't want to actually call these methods, but just ensure that they
    # have been called.
    from simmate.toolkit import Structure
    from simmate.toolkit.diffusion import MigrationHop

    mocker.patch.object(
        Structure,
        "from_dynamic",
    )
    mocker.patch.object(
        MigrationHop,
        "from_dynamic",
    )
    # -------

    example_parameters = {
        "migration_hop": None,
        "supercell_start": None,
        "supercell_end": None,
        "structures": "None; None; None",
    }

    Workflow._deserialize_parameters(**example_parameters)


# !!! These tests below are for the Prefect Executor, which is disabled at the moment

# @task
# def dummy_task_1(a):
#     return 1
# @task
# def dummy_task_2(a):
#     return 2
# @staticmethod
# def run_config(source=None, structure=None, **kwargs):
#     x = dummy_task_1(source)
#     y = dummy_task_2(structure)
#     return x + y

# # Run the workflow just like you would for the base Prefect class
# flow = DummyFlow._to_prefect_flow()
# state = flow(return_state=True, directory=tmp_path)
# assert state.is_completed()
# assert state.result() == 3

# @pytest.mark.prefect_db
# @pytest.mark.django_db
# def test_workflow_cloud(mocker, sample_structures):
#     # test cloud properties
#     deployment_id = DummyFlow.deployment_id
#     assert isinstance(deployment_id, str)
#     # we dont check the actual value bc its randomly generated
#     n = DummyFlow.nflows_submitted
#     assert isinstance(n, int)
#     # to test serialization of input parameters we grab a toolkit object
#     structure = sample_structures["C_mp-48_primitive"]
#     # Run the workflow through prefect cloud
#     flow_id = DummyFlow.run_cloud(structure=structure)
#     assert isinstance(flow_id, str)



# -*- coding: utf-8 -*-

from pathlib import Path

from pymatgen.analysis.structure_matcher import StructureMatcher

from simmate.apps.bader.workflows import PopulationAnalysis__Bader__Badelf
from simmate.apps.materials_project.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)
from simmate.database.third_parties import MatprojStructure
from simmate.engine import Workflow
from simmate.toolkit import Structure
from simmate.utilities import copy_files_from_directory


class PopulationAnalysis__VaspBader__BadelfMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density using the ELFCAR
    as a reference when partitioning.
    """

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        empty_sites: list[float] = [],
        empty_ion: str = None,
        command: str = None,
        source: dict = None,
        directory: Path = None,
        **kwargs,
    ):
        if empty_sites and empty_ion:
            raise Exception(
                "You can only specify either empty_sites or an empty_ion type. "
                "Not both."
            )

        if empty_ion:
            # Check if the input structure given already has the empty atom in
            # it. If so, we need to delete the empty atoms to get the base structure
            if empty_ion in [e.symbol for e in structure.composition.elements]:
                structure_w_empties = structure.copy()
                structure.remove_species(empty_ion)
            # Otherwise we run database queries to find a potential match
            else:
                # load the structure that contains dummy atoms in it
                structure_w_empties = get_structure_w_empties(
                    structure=structure,
                    empty_ion=empty_ion,
                )

        elif empty_sites:
            # Select an empty element that is NOT already in the structure
            for empty_ion in ["H", "He", "Lv"]:
                if empty_ion not in [e.symbol for e in structure.composition.elements]:
                    break
            structure_w_empties = structure.copy()
            for empty_coords in empty_sites:
                structure_w_empties.append(
                    species=empty_ion,
                    coords=empty_coords,  # I assume fractional coords
                )

        else:
            # otherwise no empties will be used
            # Still, select an empty element that is NOT already in the structure
            empty_ion = None
            structure_w_empties = structure.copy()

        # Run the pre-static energy calculation to generate our CHGCAR and ELFCAR
        prebadelf_dir = directory / StaticEnergy__Vasp__PrebadelfMatproj.name_full
        StaticEnergy__Vasp__PrebadelfMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=prebadelf_dir,
        ).result()

        # And run the bader analysis on the resulting chg denisty + elfcar
        badelf_dir = directory / PopulationAnalysis__Bader__Badelf.name_full
        PopulationAnalysis__Bader__Badelf.run(
            structure=structure_w_empties,
            directory=badelf_dir,
            previous_directory=prebadelf_dir,
        ).result()

        # The from_vasp_directory method that loads results into the database
        # requires the following files to be in the main directory:
        #  1. the ACF.dat
        #  2. INCAR
        #  3. vasprun.xml
        #  4. POTCAR
        #  5. CHGCAR_empty
        copy_files_from_directory(
            files_to_copy=["ACF.dat", "CHGCAR_empty"],
            directory_new=directory,
            directory_old=badelf_dir,
        )
        copy_files_from_directory(
            files_to_copy=["INCAR", "vasprun.xml", "POTCAR"],
            directory_new=directory,
            directory_old=prebadelf_dir,
        )
        # !!! I need a better way to access these files in the workup method
        # without copying them into the main dir...


# -----------------------------------------------------------------------------

# Below are extra tasks and subflows for the workflow that is defined above


class StaticEnergy__Vasp__PrebadelfMatproj(StaticEnergy__Vasp__Matproj):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis where
    the ELF is used as the reference instead of the CHGCAR.
    """

    # The key thing for bader analysis is that we need a very fine FFT mesh. Other
    # than that, it's the same as a static energy calculation.
    _incar_updates = dict(
        LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
        LELF=True,  # writes ELFCAR
        NPAR=1,  # must be set if LELF is set to True
        PREC="Single",  # ensures CHGCAR grid matches ELFCAR grid
        # Note that these set the FFT grid while the pre-Bader task sets the
        # fine FFT grid (e.g. useds NGX instead of NGXF)
        NGX__density_a=10,
        NGY__density_b=10,
        NGZ__density_c=10,
    )


def get_structure_w_empties(structure, empty_ion):
    """
    Searches the Materials Project database for a structure that contains
    the extra ion and matching host lattice, and uses it to introduce
    empty atoms into the original structure. For example, if your input
    gave a Ca2N structure and Cl ion for the template, this function will
    find Ca2NCl in the Matproj database and then use the Cl sites to
    add dummy "H" atoms into the Ca2N structure. This is particularly useful
    if you would like to analyze electrides.
    """

    # !!! It might make more sense to have this accept a database Structure object
    # and also search the same table for a match (rather than the Matproj).

    # Grab the chemical system of the structure when it includes the empty ion
    structure_dummy = structure.copy()
    # coords don't matter here bc I'm just after the chemical_system
    structure_dummy.append(empty_ion, [0, 0, 0])
    template_system = structure_dummy.composition.chemical_system

    # Go through the Materials Project database and find a structure that
    # is matching when the empty ion type is ignored.
    matcher = StructureMatcher(
        ignored_species=empty_ion,
        primitive_cell=False,  # required for the get_s2_like_s1 method
    )
    potential_matches = MatprojStructure.objects.filter(
        chemical_system=template_system
    ).all()
    for potential_match in potential_matches:
        structure_template = potential_match.to_toolkit()
        is_match = matcher.fit(structure, structure_template)
        # Once we find a match, stop searching
        if is_match:
            break

    # confirm that we successfully found a match
    if not is_match:
        raise Exception("No match found when trying to generate empty sites")

    # convert our match to an equivalent basis set as the original input
    # structure -- this ensures frac coords are matches.
    structure_template = matcher.get_s2_like_s1(structure, structure_template)

    # make a copy because we will be modifying the structure
    structure_w_empties = structure.copy()
    for site in structure_template.sites:
        if site.specie.symbol == empty_ion:
            structure_w_empties.append(empty_ion, site.frac_coords)

    # TODO: consider doing a check to ensure the empty atoms added are reasonable.
    # A simple check could be to use the SiteDistance validator and make sure
    # the hydrogen isn't too close to another atom.

    # write the new structure to file
    # filename = directory / "simmate_structure_w_empties.cif"
    # structure_w_empties.to(filename=str(filename), fmt="cif")

    return structure_w_empties



# -*- coding: utf-8 -*-

"""
This module defines the base class for all workflows in Simmate. When learning 
how use workflows, make sure you have gone through our intro 
[tutorials](https://github.com/jacksund/simmate/tree/main/tutorials). You
can then read through these guides for more features.
"""

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file


# OPTIMIZE: These imports are particularly slow because we are configuring
# django in the code below this. These two imports don't depend on the database
# module, so it may be worth isolating them for faster imports.
from .error_handler import ErrorHandler

# All imports below this point depend on the simmate.database module and therefore
# need a database connection. This line sets up the database connection so
# that models can be imported.
from simmate.database import connect


from .workflow import Workflow, workflow
from .s3_workflow import S3Workflow
from .web_api_workflow import WebApiWorkflow
from .execution import SimmateWorker as Worker



# -*- coding: utf-8 -*-

from pathlib import Path

import plotly.graph_objects as plotly_go
from plotly.subplots import make_subplots
from pymatgen.io.vasp.outputs import Vasprun

from simmate.database.base_data_types import (
    Calculation,
    Forces,
    Structure,
    Thermodynamics,
    table_column,
)
from simmate.visualization.plotting import PlotlyFigure


class Dynamics(Structure, Calculation):
    """
    Holds results from a dynamics simulations -- often referred to as a molecular
    dynamics run.

    In addition to the attributes listed, you can also access all ionic steps
    of the run via the `structures` attribute. This attribute gives a list of
    `DynamicsIonicSteps`.
    """

    class Meta:
        app_label = "workflows"

    api_filters = dict(
        temperature_start=["range"],
        temperature_end=["range"],
        time_step=["range"],
        nsteps=["range"],
    )

    temperature_start = table_column.IntegerField(blank=True, null=True)
    """
    The starting tempertature of the simulation in Kelvin.
    """

    temperature_end = table_column.IntegerField(blank=True, null=True)
    """
    The ending tempertature of the simulation in Kelvin.
    """

    time_step = table_column.FloatField(blank=True, null=True)
    """
    The time in picoseconds for each step in the simulation.
    """

    nsteps = table_column.IntegerField(blank=True, null=True)
    """
    The total number of steps in the simulation. Note, this is the maximum cutoff
    steps set. In some cases, there may be fewer steps when the simulation is 
    stopped early.
    """

    def write_output_summary(self, directory: Path):
        super().write_output_summary(directory)
        self.write_simmulation_detail_plot(directory=directory)

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun):
        raise NotImplementedError(
            "Dynamics runs cannot currently be loaded from a dir/vasprun, so"
            "input parameters such as temperature and nsteps are not loaded. "
            "This feature is still under development"
        )
        # See Relaxation.from_vasp_run as a starting point

    def update_from_directory(self, directory: Path):
        # check if we have a VASP directory
        vasprun_filename = directory / "vasprun.xml"
        if not vasprun_filename.exists():
            # raise Exception(
            #     "Only VASP output directories are supported at the moment"
            # )
            return  # just exit

        from simmate.apps.vasp.outputs import Vasprun

        vasprun = Vasprun.from_directory(directory)
        self.update_from_vasp_run(vasprun)

    def update_from_vasp_run(self, vasprun: Vasprun):
        """
        Given a Vasprun object from a finished dynamics run, this will update the
        Dynamics table entry and the corresponding DynamicsIonicStep entries.

        #### Parameters

        vasprun :
            The final Vasprun object from the dynamics run outputs.
        corrections :
            List of errors and corrections applied to during the relaxation.
        directory :
            name of the directory that relaxation was ran in. This is only used
            to reference the archive file if it's ever needed again.
        """

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key
        data = vasprun.as_dict()["output"]

        # The only other data we need to grab is the list of structures. We can
        # pull the structure for each ionic step from the vasprun class directly.
        structures = vasprun.structures

        # Now let's iterate through the ionic steps and save these to the database.
        for number, (structure, ionic_step) in enumerate(
            zip(structures, data["ionic_steps"])
        ):
            # first pull all the data together and save it to the database. We
            # are saving this to an DynamicsIonicStepStructure datatable. To access this
            # model, we look need to use "structures.model".
            structure = self.structures.model.from_toolkit(
                number=number,
                structure=structure,
                energy=ionic_step.get("e_wo_entrp", None),
                site_forces=ionic_step.get("forces", None),
                lattice_stress=ionic_step.get("stress", None),
                temperature=self._get_temperature_at_step(number),
                # simulation_time=number*self.time_step,
                dynamics_run=self,  # this links the structure to this dynamics run
            )
            structure.save()

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()

    def _get_temperature_at_step(self, step_number: int):
        return step_number * self._get_temperature_step_size() + self.temperature_start

    def _get_temperature_step_size(self):
        return (self.temperature_end - self.temperature_start) / self.nsteps


class DynamicsIonicStep(Structure, Thermodynamics, Forces):
    """
    Holds information for a single ionic step of a `Dynamics` entry.

    Each entry will map to a `Dynamics`, so you should typically access this
    data through that class. The exception to this is when you want all ionic
    steps accross many relaxations for a machine learning input.
    """

    class Meta:
        app_label = "workflows"

    archive_fields = ["number"]

    api_filters = dict(
        number=["range"],
        temperature=["range"],
    )

    number = table_column.IntegerField()
    """
    This is ionic step number for the given relaxation. This starts counting from 0.
    """

    temperature = table_column.FloatField(blank=True, null=True)
    """
    Expected temperature based on the temperature_start/end and nsteps. Note
    that in-practice some steps may not be at equilibrium temperatue. This could
    be the 1st 100 steps of a run or alternatively when the thermostat component
    is off-temperature.
    """

    dynamics_run = table_column.ForeignKey(
        Dynamics,
        on_delete=table_column.CASCADE,
        related_name="structures",
    )
    """
    All structures in this table come from dynamics run calculations, where
    there can be many structures (one for each ionic step) linked to a
    single run. This means the start structure, end structure, and
    those structure in-between are stored together here.
    Therefore, there's just a simple column stating which relaxation it
    belongs to.
    """

    # TODO:
    # simulation_time = table_column.FloatField(blank=True, null=True)
    # """
    # The total simulation time up to this ionic step (in picoseconds). This is
    # just "number*time_step", but we store this in the database for easy access.
    # """

    # TODO: Additional options from Vasprun.as_dict to consider adding
    # e_0_energy
    # e_fr_energy
    # kinetic
    # lattice kinetic
    # nosekinetic
    # nosepot


class SimmulationDetail(PlotlyFigure):
    def get_plot(result: Dynamics):
        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = result.structures.order_by("number").to_dataframe()

        # We will be making a figure that consists of 3 stacked subplots that
        # all share the x-axis of ionic_step_number
        figure = make_subplots(
            rows=4,
            cols=1,
            shared_xaxes=True,
        )

        # Generate a plot for Energy (per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["energy_per_atom"],
            ),
            row=1,
            col=1,
        )

        # Generate a plot for Forces (norm per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["site_forces_norm_per_atom"],
            ),
            row=2,
            col=1,
        )

        # Generate a plot for Stress (norm per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["lattice_stress_norm_per_atom"],
            ),
            row=3,
            col=1,
        )

        # Generate a plot for Stress (norm per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["temperature"],
            ),
            row=4,
            col=1,
        )

        # Now let's clean up some formatting and add the axes titles
        figure.update_layout(
            width=600,
            height=600,
            showlegend=False,
            xaxis3_title="Ionic Step (#)",
            yaxis_title="Energy (eV/atom)",
            yaxis2_title="Site Forces",
            yaxis3_title="Lattice Stress",
            yaxis4_title="Temperature (K)",
        )

        return figure


# register all plotting methods to the database table
for _plot in [SimmulationDetail]:
    _plot.register_to_class(Dynamics)



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.diffusion.mit import (
    Diffusion__Vasp__NebSinglePathMit,
)
from simmate.apps.materials_project.workflows.relaxation.mit import (
    Relaxation__Vasp__Mit,
)
from simmate.apps.materials_project.workflows.static_energy.mit import (
    StaticEnergy__Vasp__Mit,
)
from simmate.apps.vasp.workflows.diffusion import NebAllPathsWorkflow


class Diffusion__Vasp__NebAllPathsMit(NebAllPathsWorkflow):
    """
    Runs a full diffusion analysis on a bulk crystal structure using NEB with
    MIT settings.
    """

    bulk_relaxation_workflow = Relaxation__Vasp__Mit
    bulk_static_energy_workflow = StaticEnergy__Vasp__Mit
    single_path_workflow = Diffusion__Vasp__NebSinglePathMit



# -*- coding: utf-8 -*-

import pandas


class Selector:
    """
    The abstract base class for selecting parent individuals from a pool. This
    class should be not used directly but inheritted from.

    For adding new selection methods, DEAP is a good reference:
      https://github.com/DEAP/deap/blob/master/deap/tools/selection.py
      https://deap.readthedocs.io/en/master/api/tools.html

    Their selection methods take a list of individuals that each have a .fitness
    attribute. Instead, I pass individuals here as a pandas dataframe and have
    them specify a "fitness_column"

    Another good reference is Wikipedia
      https://en.wikipedia.org/wiki/Selection_(genetic_algorithm)

    NOTE: I assume lower fitness is better! This may change in the future because
    other selection methods throw issues with lower=better and even more issues
    when we have negative values
    """

    @classmethod
    def select(
        cls,
        nselect: int,
        individuals: pandas.DataFrame,
        fitness_column: str,
    ):
        # The select method should take in two arguments:
        # individuals = a pandas dataframe with columns "id" and "fitness"
        # n = the number of individuals that we want to choose from those fitnesses
        # The select method should then return a list/tuple of the indicies
        # of the fitnesses that were selected
        raise NotImplementedError(
            "You must add a custom select method for your Selector subclass"
        )

    @classmethod
    @property
    def name(cls):
        """
        A nice string name for the selector. By default it just returns the name
        of this class.
        """
        return cls.__name__

    @classmethod
    def select_from_datatable(
        cls,
        nselect: int,
        datatable,  # Queryset
        fitness_column: str = None,
        query_limit: str = None,
    ):
        # our selectors just require a dataframe where we specify the fitness
        # column. So we query our individuals database to give this as an input.
        datatable_cleaned = datatable
        if fitness_column:
            datatable_cleaned = datatable_cleaned.order_by(fitness_column)
        if query_limit:
            datatable_cleaned = datatable_cleaned[:query_limit]

        individuals_df = datatable_cleaned.to_dataframe()

        # From these individuals, select our parent structures
        parents_df = cls.select(
            nselect=nselect,
            individuals=individuals_df,
            fitness_column=fitness_column,
        )

        # grab the id column of the parents and convert it to a list
        parent_ids = parents_df.id.values.tolist()

        # Now lets grab these structures from our database and convert them
        # to a list of pymatgen structures.
        # We have to make separate queries for this instead of doing "id__in=parent_ids".
        # This is because (1) order may be important and (2) we may have duplicate
        # entries. For example, a hereditary mutation can request parent ids of [123,123]
        # in which case we want to give the same input structure twice!
        parent_structures = [
            datatable.only("structure").get(id=parent_id).to_toolkit()
            for parent_id in parent_ids
        ]

        # When there's only one structure selected we return the structure and
        # id independently -- not within a list
        if nselect == 1:
            parent_ids = parent_ids[0]
            parent_structures = parent_structures[0]

        # for record keeping, we also want to return the ids for each structure
        return parent_ids, parent_structures



# -*- coding: utf-8 -*-

"""
This is where you add any new database tables. 

This file is named "models.py" because that is what Django expects! In their 
terminology, a django 'model' is the same thing as a 'database table'.

See our online documentation for more information.
"""

from simmate.database.base_data_types import (
    Calculation,
    DatabaseTable,
    Structure,
    table_column,
)


# OPTION 1:
# We can make a DatabaseTable from scratch. This table will only have the
# columns we define, plus a few default ones: id, created_at, update_at, and source
class MyCustomTable1(DatabaseTable):
    custom_column_01 = table_column.IntegerField()
    custom_column_02 = table_column.FloatField(null=True, blank=True)


# OPTION 2:
# Mix-ins let you build out custom tables with common types of data. These will
# automatically build out your table with many different columns. Many
# also have a `from_toolkit` method that makes populating your database super easy
class MyCustomTable2(Structure, Calculation):
    input_01 = table_column.FloatField(null=True, blank=True)
    input_02 = table_column.BooleanField(null=True, blank=True)
    output_01 = table_column.FloatField(null=True, blank=True)
    output_02 = table_column.BooleanField(null=True, blank=True)


# OPTION 3: (advanced users only)
# You can build a table using django instead of simmate
# https://docs.djangoproject.com/



# -*- coding: utf-8 -*-

"""
Simmate Configuration
---------------------

This module sets up Simmate and other programs it depends on. It also includes an example config directory and example custom project. These example directories are only available through [github](https://github.com/jacksund/simmate/tree/main/src/simmate/configuration) and not the API documentation.
"""

from .load_settings import settings



# -*- coding: utf-8 -*-

from simmate.apps.badelf.workflows.badelf import BadElf__Badelf__Badelf
from simmate.apps.badelf.workflows.base import VaspBadElfBase
from simmate.apps.warren_lab.workflows.badelf.prebadelf_dft import (
    StaticEnergy__Vasp__WarrenLabPrebadelfPbesol,
)


class BadElf__Badelf__BadelfPbesol(VaspBadElfBase):
    """
    Runs a static energy calculation using an extra-fine FFT grid using vasp
    and then carries out Badelf and Bader analysis on the resulting charge density.
    Uses the Warren lab settings Pbesol settings.
    """

    static_energy_workflow = StaticEnergy__Vasp__WarrenLabPrebadelfPbesol
    badelf_workflow = BadElf__Badelf__Badelf



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Structure, table_column


class JarvisStructure(Structure):
    """
    Crystal structures from the [JARVIS](https://jarvis.nist.gov/) database.

    Currently, this table only stores strucure and reported energy above hull.
    The calculated energy is not reported, so the Thermodynamics mixin is not used.
    """

    class Meta:
        app_label = "data_explorer"

    archive_fields = ["energy_above_hull"]
    api_filters = dict(
        energy_above_hull=["range"],
    )
    source = "JARVIS"
    source_long = "Joint Automated Repository for Various Integrated Simulations"
    homepage = "https://jarvis.nist.gov/"
    source_doi = "https://doi.org/10.1038/s41524-020-00440-1"
    remote_archive_link = "https://archives.simmate.org/JarvisStructure-2023-07-07.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "jvasp-12345")
    """

    # TODO: contact their team to ask about reporting energy instead. That way
    # we can use the Thermodynamics mixin instead of manually listing this.
    energy_above_hull = table_column.FloatField(blank=True, null=True)
    """
    The energy above hull, as reported by the JARVIS database (no units given)
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the JARVIS website.
        """
        # All JARVIS structures have their data mapped to a URL in the same way
        # ex: https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/JVASP-1234.xml
        # we store the id as "jvasp-123" so we need to convert this to uppercase
        id = self.id.upper()
        return f"https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/{id}.xml"

    # -------------------------------------------------------------------------

    @classmethod
    def _load_all_structures(cls):
        """
        Only use this function if you are part of the Simmate dev team!
        Users should instead access data via the load_remote_archive method.

        This method pulls JARVIS data into the Simmate database.

        JARVIS has a python package "jarvis-tools" that let's us pull some of
        their database dumps. For instructions on how to do this, they provided
        [this link](https://colab.research.google.com/github/knc6/jarvis-tools-notebooks/blob/master/jarvis-tools-notebooks/Get_JARVIS_DFT_final_structures_in_ASE_or_Pymatgen_format.ipynb)

        Alternatively, we could manually download
        [their database json files](https://jarvis-materials-design.github.io/dbdocs/thedownloads/). We specifically look at the "3D-materials curated data".
        """

        from rich.progress import track

        from simmate.toolkit import Structure as ToolkitStructure

        # Jarvis is not a dependency of simmate, so make sure you install it
        # before using this method
        try:
            from jarvis.db.figshare import data as jarvis_helper
        except:
            raise ModuleNotFoundError(
                "You must install jarvis with `conda install -c conda-forge jarvis-tools`"
            )

        # Load all of the 3D data from JARVIS. This gives a list of dictionaries
        # TODO: In the future, we can add other datasets like the 2D dataset.
        data = jarvis_helper("dft_3d")

        # Now iterate through all the data -- which is a list of dictionaries.
        # We convert the data into a pymatgen object and sanitize it before
        # saving to the Simmate database
        db_objects = []
        for entry in track(data):
            # The structure is in the atoms field as a dictionary. We pull
            # this data out and convert it to a pymatgen Structure object
            structure = ToolkitStructure(
                lattice=entry["atoms"]["lattice_mat"],
                species=entry["atoms"]["elements"],
                coords=entry["atoms"]["coords"],
                coords_are_cartesian=entry["atoms"]["cartesian"],
            )

            # now convert the entry to a database object
            structure_db = cls.from_toolkit(
                id=entry["jid"].lower(),
                structure=structure,
                energy_above_hull=entry["ehull"] if entry["ehull"] != "na" else None,
            )

            db_objects.append(structure_db)

        # and save it to our database
        cls.objects.bulk_create(
            db_objects,
            batch_size=15000,
            ignore_conflicts=True,
        )



# Generated by Django 4.2.2 on 2023-07-21 16:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="FingerprintPool",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("method", models.JSONField(blank=True, null=True)),
                ("init_kwargs", models.JSONField(blank=True, null=True)),
                (
                    "database_table",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Spacegroup",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("number", models.IntegerField(primary_key=True, serialize=False)),
                ("symbol", models.CharField(max_length=15)),
                ("crystal_system", models.CharField(max_length=15)),
                ("point_group", models.CharField(max_length=15)),
            ],
        ),
        migrations.CreateModel(
            name="Fingerprint",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("database_id", models.IntegerField(blank=True, null=True)),
                ("fingerprint", models.JSONField(blank=True, null=True)),
                (
                    "pool",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fingerprints",
                        to="core_components.fingerprintpool",
                    ),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-

from .hse import Relaxation__Vasp__WarrenLabHse
from .hsesol import Relaxation__Vasp__WarrenLabHsesol
from .pbe import Relaxation__Vasp__WarrenLabPbe
from .pbe_metal import Relaxation__Vasp__WarrenLabPbeMetal
from .pbesol import Relaxation__Vasp__WarrenLabPbesol
from .scan import Relaxation__Vasp__WarrenLabScan



# -*- coding: utf-8 -*-
import logging
import math
from functools import cached_property
from pathlib import Path

import numpy as np
from pymatgen.analysis.local_env import CrystalNN
from pymatgen.core.periodic_table import _pt_data as pt_data
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.apps.badelf.core.grid import Grid
from simmate.apps.badelf.core.partitioning import PartitioningToolkit
from simmate.apps.bader.outputs import ACF
from simmate.workflows.utilities import get_workflow


class ElectrideFinder:
    """
    A class for finding electride sites from an ELFCAR.

    Args:
        grid : Grid
            A BadELF app Grid instance made from an ELFCAR.
    """

    def __init__(
        self,
        grid: Grid,
        directory: Path,
    ):
        self.grid = grid.copy()
        self.directory = directory

    @cached_property
    def local_maxima(self):
        """
        The local maxima in a 3D numpy array
        """
        return self.find_local_maxima()

    def find_local_maxima(self, neighborhood_size=1, threshold=None):
        """
        Find local maxima in a 3D numpy array.

        Args:
        - neighborhood_size: Size of the neighborhood for finding local maxima
        - threshold: Threshold for considering a point as a local maximum

        Returns:
        - List of tuples containing the coordinates of local maxima
        """
        grid = self.grid.copy()
        grid.regrid(desired_resolution=1000)
        elf_data = grid.total
        # Get padded data so that we can look at voxels at the edges
        padded_elf_data = np.pad(elf_data, neighborhood_size, mode="wrap")
        maxima_cart_coords = []
        maxima_values = []

        # Look across each voxel in the structure.
        for z in range(neighborhood_size, padded_elf_data.shape[0] - neighborhood_size):
            for y in range(
                neighborhood_size, padded_elf_data.shape[1] - neighborhood_size
            ):
                for x in range(
                    neighborhood_size, padded_elf_data.shape[2] - neighborhood_size
                ):
                    # Get a section of the dataframe around the voxel
                    neighborhood = padded_elf_data[
                        z - neighborhood_size : z + neighborhood_size + 1,
                        y - neighborhood_size : y + neighborhood_size + 1,
                        x - neighborhood_size : x + neighborhood_size + 1,
                    ]
                    # Get the max value in the neighborhood
                    max_value = np.max(neighborhood)
                    z_orig = z - neighborhood_size
                    y_orig = y - neighborhood_size
                    x_orig = x - neighborhood_size
                    # If the maximum value is at the voxel we're looking at, this
                    # is a maximum and we add the cartesian coordinates and value
                    # to our list
                    if elf_data[z_orig, y_orig, x_orig] == max_value and (
                        threshold is None or max_value > threshold
                    ):
                        maxima_voxel_coord = (z_orig + 1, y_orig + 1, x_orig + 1)
                        maxima_cart_coord = grid.get_cart_coords_from_vox(
                            maxima_voxel_coord
                        )
                        maxima_cart_coords.append(maxima_cart_coord)
                        maxima_values.append(elf_data[z_orig, y_orig, x_orig])

        return maxima_cart_coords, maxima_values

    @staticmethod
    def to_number_from_roman_numeral(roman_num):
        lookup = {
            "X": 10,
            "V": 5,
            "I": 1,
        }
        total = 0
        i = 0
        while i < len(roman_num):
            # If this is the subtractive case.
            if (
                i + 1 < len(roman_num)
                and lookup[roman_num[i]] < lookup[roman_num[i + 1]]
            ):
                total += lookup[roman_num[i + 1]] - lookup[roman_num[i]]
                i += 2
            # Else this is NOT the subtractive case.
            else:
                total += lookup[roman_num[i]]
                i += 1

        return total

    @cached_property
    def shannon_radii_estimate(self):
        """
        Gets the shannon crystal radii for each atom in the structure. The
        oxidation states are estimated using traditional bader.
        """
        directory = self.directory
        structure = self.grid.structure
        badelf_workflow = get_workflow("population-analysis.bader.bader-dev")
        badelf_workflow.run(
            directory=directory,
        )

        # Get oxidation states for each atom as a list
        results_dataframe, extra_data = ACF(directory)
        oxidation_states = results_dataframe.oxidation_state.values
        oxidation_states = np.round(oxidation_states)
        # add method to get coordination environments. This already exists in
        # badelftoolkit but I don't want to repeat it for no reason.
        cnn = CrystalNN()
        shannon_radii = {}
        for site_index, site in enumerate(structure):
            # We need the shannon radii data for the atom we're investigating.
            # We get this from pymatgen's pt_data dictionary
            site_species = structure[site_index].species_string
            species_dict = pt_data[site_species]["Shannon radii"]

            # Now we need the oxidation state and coordination environment of
            # our atom of interest in order to get the closest matching shannon
            # radius. For electrides we'll often get estimated oxidation states
            # that are far from accurate, so we get the closest available radius
            oxidation_state = oxidation_states[site_index]
            coord_env = cnn.get_cn(structure, site_index)

            available_ox_state = species_dict.keys()
            # We find the closest oxidation state to the one we got from bader.
            # The min function here is taking in the available oxidation values.
            # It then looks at every available oxidation state (x) and calculates
            # the difference between it and the bader ox value and stores it as
            # a key. It then returns the available oxidation state with the lowest
            # key value.
            closest_ox_state = min(
                available_ox_state, key=lambda x: abs(int(x) - oxidation_state)
            )

            available_coord_envs = species_dict[closest_ox_state].keys()
            # We similarly find the closest coordination env. These are stored
            # as roman numerals in the pymatgen pt_data so they must be converted
            # to integers
            closest_coord_env = min(
                available_coord_envs,
                key=lambda x: abs(self.to_number_from_roman_numeral(x) - coord_env),
            )

            # We now get the closest shannon prewitt crystal radius
            shannon_radius = species_dict[closest_ox_state][closest_coord_env][""][
                "crystal_radius"
            ]
            shannon_radii[site_index] = shannon_radius
        return shannon_radii

    @cached_property
    def elf_radii(self):
        """
        Gets the elf radius of each atom in the structure
        """
        grid = self.grid.copy()
        structure = grid.structure.copy()
        # We don't want to find the radius for every atom in the structure if
        # there is symmetry. We instead find the radius for each unique atom
        symmetry_dataset = SpacegroupAnalyzer(structure).get_symmetry_dataset()
        equivalent_atoms = symmetry_dataset["equivalent_atoms"]

        # We now have an array of the same length as the structure. Each value
        # in the array is an int pointing to the first equivalent atom in the
        # structure if it exists (or itself if its the earliest equivalent one).
        # Now we want to get the ELF radius for each unique atom. We iterate over
        # each unique atom to get its radius and add it to a dictionary.
        unique_atoms = np.unique(equivalent_atoms)
        unique_atom_radii = {}
        for atom in unique_atoms:
            radius = PartitioningToolkit(grid).get_elf_ionic_radius(
                site_index=atom,
                structure=structure,
            )
            unique_atom_radii[atom] = radius

        # We now construct a dictionary for all of the atoms in the structure
        # that provides its radius
        all_atom_radii = {}
        for site_index, site in enumerate(structure):
            equiv_atom = equivalent_atoms[site_index]
            radius = unique_atom_radii[equiv_atom]
            all_atom_radii[site_index] = radius

        return all_atom_radii

    def get_ionic_radii(
        self,
        method="elf",
    ):
        """
        Gets the radius for each atom in the structure.

        Parameters
        ----------

        method : str, optional
            The method used to find the radii. Options are "elf" or "shannon".
            elf will use the provided grid to find the radius of each atom and
            shannon will use bader oxidation states and the closest available
            tabulated shannon crystal radius.
        """
        if method == "elf":
            atom_radii = self.elf_radii
        elif method == "shannon":
            atom_radii = self.shannon_radii_estimate
        else:
            raise ValueError(
                """Selected method for finding ionic radii does not exist. Available
                methods are "elf" and "shannon"
                """
            )
        return atom_radii

    def get_electride_structure(
        self,
        electride_finder_cutoff: float = 0.5,
        min_electride_radius: float = 0.9,
        atom_radius_method: str = "elf",
        remove_old_electrides: bool = False,
        local_maxima_coords: list = None,
        local_maxima_values: list = None,
    ):
        #!!! The min_electride_radius is based off of fluoride, but it feels
        # arbitrary. Is there a better way?
        """
        Finds the electrides in a structure using an ELF grid.

        Parameters
        ----------

        electride_finder_cutoff : float, optional
            The minimum elf value at the site to be considered a possible
            electride. The default is 0.5.
        min_electride_radius : float, optional
            The minimum elf radius around the maximum for it to be considered
            an electride. The default is 0.9 which is somewhat arbitrarily chosen.
            An goold alternative is 1.19 which is the average radius of fluoride
            in a 6 coordination environment.
        atom_radius_method : str, optional
            The method used to find the radii. Options are "elf" or "shannon".
            elf will use the provided grid to find the radius of each atom and
            shannon will use bader oxidation states and the closest available
            tabulated shannon crystal radius.
        local_maxima_coords : list, optional
            The coordinates of all local maxima in the grid. This will be found
            automatically if not set.
        local_maxima_values : list, optional
            The values at the local maxima. This will be found automatically if
            not set.
        remove_old_electrides : bool, optional
            Whether or not to remove any other electrides already placed in
            the system. It is generally recommended that structures without
            electrides are provided.

        Returns
        -------
        A structure object with the found electride sites labeled with "He"
        dummy atoms.
        """
        logging.info("Finding electride sites")
        # Get the coordinates and values of each local maximum in the grid
        grid = self.grid.copy()
        if local_maxima_coords is None:
            local_maxima_coords, local_maxima_values = self.local_maxima

        # If there are He atoms that have already been placed in the structure
        # that the user wants to remove, remove them now.
        structure = grid.structure.copy()
        if "He" in structure.symbol_set and remove_old_electrides:
            old_electride_sites = structure.indices_from_symbol("He")
            structure.remove_sites(old_electride_sites)
        elif "He" in structure.symbol_set and not remove_old_electrides:
            logging.warning(
                """
                  Dummy atoms were found already in the structure. No new
                  electride sites will be added. To have the algorithm search
                  for electride sites, remove all He dummy atoms in the structure.
                  """
            )
            return structure

        # Create a list to store the electride coords.
        cnn = CrystalNN()
        electride_coords = []
        # get the estimated shannon radii of each of the sites in the structure.
        # These will be used to estimate electride radii
        shannon_radii = self.get_ionic_radii(atom_radius_method)
        for i, maximum_coords in enumerate(local_maxima_coords):
            # Check if the elf value is below the cutoff
            if local_maxima_values[i] < electride_finder_cutoff:
                continue

            # Create a dummy structure. Add a dummy at the local maximum that
            # was found
            electride_structure = structure.copy()
            electride_structure.append("He", maximum_coords, coords_are_cartesian=True)
            # Get the nearest neighbors of the electride. cnn will return a dict
            # with multiple keys. These might be different methods of finding neighbors?
            # I pick the one with the most neighbors
            _, _, electride_neighbors = cnn.get_nn_data(electride_structure, n=-1)
            most_neighbors = max(electride_neighbors)
            electride_neighbors = electride_neighbors[most_neighbors]

            # get the distance from the potential electride site to the edge
            # of the atoms shannon radius. This is the maximum radius that the
            # electride can have.
            max_electride_radii = []
            for neighbor in electride_neighbors:
                neighbor_index = neighbor["site_index"]
                neighbor_shann_rad = shannon_radii[neighbor_index]
                if neighbor["site"].species_string != "He":
                    distance_to_atom = math.dist(
                        maximum_coords, neighbor["site"].coords
                    )
                    max_radius = distance_to_atom - neighbor_shann_rad
                    max_electride_radii.append(max_radius)

            # check if minimum distance is too close to an atom:

            if min(max_electride_radii) < min_electride_radius:
                continue

            # The following code is another method for finding the radius of the
            # electrides using ELF. It works well in many situations but is
            # less lenient than using the atoms shannon radii. It is kept here
            # in case it becomes more desirable in the future.
            # # get distances to min along lines to closest atoms
            # electride_grid = grid.copy()
            # electride_grid.structure = electride_structure
            # electride_radius = PartitioningToolkit(electride_grid).get_elf_ionic_radius(
            #     site_index=len(electride_structure) - 1,
            #     structure=electride_structure,
            # )
            # if electride_radius < min_electride_radius:
            #     continue

            # If the loop is still going, we consider this site an electride. We
            # add it to the list of electride sites.
            electride_coords.append(maximum_coords)
            # electride_coordinations.append(cnn.get_cn(electride_structure, n=-1))

        # Add our potential electride sites to our structure
        electride_structure = structure.copy()
        for coord in electride_coords:
            electride_structure.append("He", coord, coords_are_cartesian=True)

        # Often the algorithm will find several electride sites right next to
        # eachother. This can be due to voxelation or because of oddly shaped
        # electrides. We want to combine these into one electride site.
        electride_structure.merge_sites(tol=0.5, mode="average")

        # The above pymatgen method often reorders the atoms in the structure which
        # causes issues down the line. We want to resort the atoms into their
        # original order with electrides at the end. To do this, we create another
        # copy of the original structure and add all of the final electride sites
        final_electride_structure = structure.copy()
        electride_sites = electride_structure.indices_from_symbol("He")
        if len(electride_sites) > 0:
            logging.info(f"{len(electride_sites)} electride sites found")
            for site in electride_sites:
                coord = electride_structure[site].coords
                final_electride_structure.append("He", coord, coords_are_cartesian=True)
        else:
            logging.info("No electride sites found")
        return final_electride_structure



# -*- coding: utf-8 -*-

from .badelf import BadElfToolkit
from .electride_finder import ElectrideFinder
from .grid import Grid
from .partitioning import PartitioningToolkit
from .voxel_assignment import VoxelAssignmentToolkit



# -*- coding: utf-8 -*-

from .prepop_relaxation_static_hse_hse import (
    Nested__Vasp__WarrenLabRelaxationStaticHseHse,
    Relaxation__Vasp__WarrenLabHseWithWavecar,
)
from .prepop_relaxation_static_pbesol_hse import (
    Nested__Vasp__WarrenLabRelaxationStaticPbeHse,
    Relaxation__Vasp__WarrenLabPbeWithWavecar,
)
from .prepop_relaxation_static_pbesol_pbesol import (
    Nested__Vasp__WarrenLabRelaxationStaticPbePbe,
)



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.engine import Workflow
from simmate.toolkit import Structure
from simmate.toolkit.diffusion import MigrationHop, MigrationImages
from simmate.toolkit.diffusion.utilities import clean_start_end_images


class SinglePathWorkflow(Workflow):
    """
    Runs a full diffusion analysis on a single migration hop using NEB.

    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you.

    For a given migration path, supercell structures are generated for
    the start and end points of the migration, where vacancy-based diffsion
    is used. These supercells are relaxed using the relaxation/neb-endpoint
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images

    Currently, this workflow will not run through the command-line, as we
    have not implemented a file format for MigrationHop's yet.
    """

    endpoint_relaxation_workflow: Workflow = None
    endpoint_energy_workflow: Workflow = None
    from_images_workflow: Workflow = None

    # Oddly enough, the from_images_workflow and this workflow share a table
    # entry, so nothing needs to be done for working up results. See
    # how we pass run_id=run_id below.
    update_database_from_results = False

    @classmethod
    def run_config(
        cls,
        migration_hop: MigrationHop,
        directory: Path = None,
        source: dict = None,
        command: str = None,
        diffusion_analysis_id: int = None,
        is_restart: bool = False,
        relax_endpoints: bool = True,
        # parameters for supercell and image generation
        nimages: int = 5,
        min_atoms: int = 80,
        max_atoms: int = 240,
        min_length: float = 10,
        run_id: str = None,
        **kwargs,
    ):
        # get the supercell endpoint structures
        supercell_start, supercell_end, _ = migration_hop.get_sc_structures(
            min_atoms=min_atoms,
            max_atoms=max_atoms,
            min_length=min_length,
            vac_mode=True,
        )

        # BUG-CHECK to ensure sites are in proper order
        # Consider moving this check into the get_sc_structures method.
        supercell_start, supercell_end = clean_start_end_images(
            supercell_start,
            supercell_end,
        )
        ## end bugcheck

        if relax_endpoints:
            # Relax the starting supercell structure
            endpoint_start_state = cls.endpoint_relaxation_workflow.run(
                structure=supercell_start,
                command=command,  # subcommands["command_supercell"]
                directory=directory
                / f"{cls.endpoint_relaxation_workflow.name_full}.start",
                is_restart=is_restart,
            )

            # Relax the ending supercell structure
            endpoint_end_state = cls.endpoint_relaxation_workflow.run(
                structure=supercell_end,
                command=command,  # subcommands["command_supercell"]
                directory=directory
                / f"{cls.endpoint_relaxation_workflow.name_full}.end",
                is_restart=is_restart,
            )

            # wait for the endpoint relaxations to finish and use them to
            # update the structures we are using
            supercell_start = endpoint_start_state.result()
            supercell_end = endpoint_end_state.result()

        # Run static-energy calculations for the endpoints
        start_energy = cls.endpoint_energy_workflow.run(
            structure=supercell_start,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{cls.endpoint_energy_workflow.name_full}.start",
            is_restart=is_restart,
        )
        end_energy = cls.endpoint_energy_workflow.run(
            structure=supercell_end,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{cls.endpoint_energy_workflow.name_full}.end",
            is_restart=is_restart,
        )
        supercell_start = start_energy.result()
        supercell_end = end_energy.result()

        # use the endpoints to generate intermediate images
        # Make sure we have toolkit objects, and if not, convert them
        supercell_start_cleaned = Structure.from_dynamic(supercell_start)
        supercell_end_cleaned = Structure.from_dynamic(supercell_end)
        # Then generate the images
        images = MigrationImages.from_endpoints(
            structure_start=supercell_start_cleaned,
            structure_end=supercell_end_cleaned,
            nimages=nimages,
            # TODO: Add extra IDPP relaxation parameters and an option to disable it
            # relaxation to the diffusing species here.
            # species=[migration_hop.isite.specie.symbol],
            # sort_tol=0,  # sites in our structures should already be sortedw
            # idpp_relax=False,
        )

        # Run NEB on this set of images
        neb_state = cls.from_images_workflow.run(
            migration_images=images,
            command=command,  # subcommands["command_neb"]
            source=source,
            directory=directory,
            diffusion_analysis_id=diffusion_analysis_id,
            is_restart=is_restart,
            # Run id is very important here as it tells the underlying
            # workflow that it doesn't need to create a new database object
            # during registration -- as it will use the one that was registered
            # when this workflow started. This is also why we have a dummy
            # `update_database_from_results` method below
            run_id=run_id,
        )
        neb_images = neb_state.result()



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class StaticEnergy__Vasp__Matproj(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MPStaticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPStaticSet).

    Runs a VASP static energy calculation using Materials Project settings.

    This is identical to relaxation/Matproj, but just a single ionic step.
    """

    _incar_updates = dict(
        IBRION=-1,  # (optional) locks everything between ionic steps
        NSW=0,  # this is the main static energy setting
        LAECHG=True,
        LCHARG=True,
        LORBIT=11,
        LVHAR=True,
        LWAVE=False,
        ALGO="Normal",  # was "Fast" before
    )



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj_hse import (
    Relaxation__Vasp__MatprojHse,
)


class Relaxation__Vasp__MatprojHsesol(Relaxation__Vasp__MatprojHse):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet).

    Runs a VASP relaxation calculation using Materials Project HSE settings with
    an additional HSEsol setting.
    """

    description_doc_short = "based on pymatgen's MPHSERelaxSet"

    # Tell VASP to use PBEsol instead of base PBE
    _incar_updates = dict(GGA="PS")



# -*- coding: utf-8 -*-

import pytest

from simmate.conftest import copy_test_files
from simmate.engine import Workflow
from simmate.workflows.utilities import (
    get_all_workflow_names,
    get_all_workflow_types,
    get_apps_by_type,
    get_unique_parameters,
    get_workflow,
    get_workflow_names_by_type,
    load_results_from_directories,
)


def test_get_workflow_types():
    assert get_all_workflow_types() == [
        # "cluster-expansion",
        # "bad-elf-analysis",
        "customized",
        "diffusion",
        "dynamics",
        "electronic-structure",
        # "nested",
        "population-analysis",
        "relaxation",
        "restart",
        "static-energy",
        "structure-prediction",
    ]


def test_list_of_all_workflows():
    assert get_all_workflow_names() == [
        # "cluster-expansion.clease.bulk-structure",
        # "bad-elf-analysis.badelf.badelf",
        # "bad-elf-analysis.badelf.badelf-hse",
        # "bad-elf-analysis.badelf.badelf-pbesol",
        "customized.toolkit.user-config",
        "diffusion.vasp.neb-all-paths-mit",
        # "diffusion.vasp.neb-all-paths-warren-lab",
        # "diffusion.vasp.neb-all-paths-warren-lab-quick",
        "diffusion.vasp.neb-from-endpoints-mit",
        "diffusion.vasp.neb-from-images-mit",
        "diffusion.vasp.neb-from-images-mvl-ci",
        "diffusion.vasp.neb-single-path-mit",
        "dynamics.vasp.matproj",
        "dynamics.vasp.mit",
        "dynamics.vasp.mvl-npt",
        "electronic-structure.vasp.matproj-band-structure",
        "electronic-structure.vasp.matproj-band-structure-hse",
        "electronic-structure.vasp.matproj-density-of-states",
        "electronic-structure.vasp.matproj-density-of-states-hse",
        "electronic-structure.vasp.matproj-full",
        "electronic-structure.vasp.matproj-hse-full",
        # "nested.vasp.warren-lab-relaxation-static-hse-hse",
        # "nested.vasp.warren-lab-relaxation-static-pbe-hse",
        # "nested.vasp.warren-lab-relaxation-static-pbe-pbe",
        "population-analysis.bader.badelf",
        "population-analysis.bader.bader",
        "population-analysis.bader.bader-dev",
        "population-analysis.bader.combine-chgcars",
        "population-analysis.vasp-bader.badelf-matproj",
        "population-analysis.vasp-bader.bader-matproj",
        "population-analysis.vasp.elf-matproj",
        # "relaxation.vasp.cluster-high-q",
        # "relaxation.vasp.cluster-low-q",
        "relaxation.vasp.matproj",
        "relaxation.vasp.matproj-hse",
        "relaxation.vasp.matproj-hsesol",
        "relaxation.vasp.matproj-metal",
        "relaxation.vasp.matproj-pbesol",
        "relaxation.vasp.matproj-scan",
        "relaxation.vasp.mit",
        "relaxation.vasp.mvl-grainboundary",
        "relaxation.vasp.mvl-neb-endpoint",
        "relaxation.vasp.mvl-slab",
        "relaxation.vasp.quality00",
        "relaxation.vasp.quality01",
        "relaxation.vasp.quality02",
        "relaxation.vasp.quality03",
        "relaxation.vasp.quality04",
        "relaxation.vasp.staged",
        # "relaxation.vasp.warren-lab-hse",
        # "relaxation.vasp.warren-lab-hse-with-wavecar",
        # "relaxation.vasp.warren-lab-hsesol",
        # "relaxation.vasp.warren-lab-pbe",
        # "relaxation.vasp.warren-lab-pbe-metal",
        # "relaxation.vasp.warren-lab-pbe-with-wavecar",
        # "relaxation.vasp.warren-lab-pbesol",
        # "relaxation.vasp.warren-lab-scan",
        # "relaxation.vasp.staged-cluster",
        "restart.toolkit.automatic",
        # "static-energy.vasp.cluster-high-qe",
        "static-energy.quantum-espresso.quality00",
        "static-energy.vasp.matproj",
        "static-energy.vasp.matproj-hse",
        "static-energy.vasp.matproj-hsesol",
        "static-energy.vasp.matproj-pbesol",
        "static-energy.vasp.matproj-scan",
        "static-energy.vasp.mit",
        "static-energy.vasp.mvl-neb-endpoint",
        "static-energy.vasp.prebadelf-matproj",
        "static-energy.vasp.prebader-matproj",
        "static-energy.vasp.quality04",
        # "static-energy.vasp.warren-lab-hse",
        # "static-energy.vasp.warren-lab-hsesol",
        # "static-energy.vasp.warren-lab-pbe",
        # "static-energy.vasp.warren-lab-pbe-metal",
        # "static-energy.vasp.warren-lab-pbesol",
        # "static-energy.vasp.warren-lab-prebadelf-hse",
        # "static-energy.vasp.warren-lab-prebadelf-pbesol",
        # "static-energy.vasp.warren-lab-scan",
        "structure-prediction.toolkit.chemical-system",
        "structure-prediction.toolkit.fixed-composition",
        "structure-prediction.toolkit.new-individual",
        "structure-prediction.toolkit.variable-nsites-composition",
    ]


def test_list_of_apps_by_type():
    assert get_apps_by_type("static-energy") == ["quantum-espresso", "vasp"]

    with pytest.raises(TypeError):
        get_apps_by_type("non-existant-type")


def test_list_of_workflows_by_type():
    assert get_workflow_names_by_type("static-energy") == [
        "static-energy.quantum-espresso.quality00",
        # "static-energy.vasp.cluster-high-qe",
        "static-energy.vasp.matproj",
        "static-energy.vasp.matproj-hse",
        "static-energy.vasp.matproj-hsesol",
        "static-energy.vasp.matproj-pbesol",
        "static-energy.vasp.matproj-scan",
        "static-energy.vasp.mit",
        "static-energy.vasp.mvl-neb-endpoint",
        "static-energy.vasp.prebadelf-matproj",
        "static-energy.vasp.prebader-matproj",
        "static-energy.vasp.quality04",
        # "static-energy.vasp.warren-lab-hse",
        # "static-energy.vasp.warren-lab-hsesol",
        # "static-energy.vasp.warren-lab-pbe",
        # "static-energy.vasp.warren-lab-pbe-metal",
        # "static-energy.vasp.warren-lab-pbesol",
        # "static-energy.vasp.warren-lab-prebadelf-hse",
        # "static-energy.vasp.warren-lab-prebadelf-pbesol",
        # "static-energy.vasp.warren-lab-scan",
    ]

    assert (
        get_workflow_names_by_type(
            "static-energy",
            app_name="non-existant",
        )
        == []
    )

    with pytest.raises(TypeError):
        get_workflow_names_by_type("non-existant-type")


def test_get_workflow():
    from simmate.apps.materials_project.workflows import (
        StaticEnergy__Vasp__Matproj as workflow,
    )

    assert get_workflow("static-energy.vasp.matproj") == workflow


# This is for the test below on custom workflows
WORKFLOW_SCRIPT = """
from simmate.engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):

    use_database = False  # we don't have a database table yet

    @staticmethod
    def run_config(**kwargs):
        print("This workflow doesn't do much")
        return 42

my_workflow = Example__Python__MyFavoriteSettings
"""


def test_get_custom_workflow(tmp_path):
    script_name = tmp_path / "my_script.py"

    with script_name.open("w") as file:
        file.write(WORKFLOW_SCRIPT)

    workflow = get_workflow(f"{script_name}:my_workflow")

    assert issubclass(workflow, Workflow)


def test_get_unique_paramters():
    assert get_unique_parameters() == [
        "angle_tolerance",
        "atoms_to_print",
        "best_survival_cutoff",
        "charge_file",
        "chemical_system",
        "command",
        "composition",
        "compress_output",
        "convergence_cutoff",
        "diffusion_analysis_id",
        "directory",
        "directory_new",
        "directory_old",
        "empty_ion",
        "empty_sites",
        "fitness_field",
        "input_parameters",
        "is_restart",
        "max_atoms",
        "max_path_length",
        "max_stoich_factor",
        "max_structures",
        "max_supercell_atoms",
        "migrating_specie",
        "migration_hop",
        "migration_images",
        "min_atoms",
        "min_length",
        "min_structures_exact",
        "min_supercell_atoms",
        "min_supercell_vector_lengths",
        "nfirst_generation",
        "nimages",
        "nsteadystate",
        "nsteps",
        "partitioning_file",
        "percolation_mode",
        "relax_bulk",
        "relax_endpoints",
        "run_id",
        "search_id",
        "selector_kwargs",
        "selector_name",
        "singleshot_sources",
        "sleep_step",
        "source",
        "species_to_print",
        "standardize_structure",
        "steadystate_source_id",
        "steadystate_sources",
        "structure",
        "subworkflow_kwargs",
        "subworkflow_name",
        "supercell_end",
        "supercell_start",
        "symmetry_precision",
        "temperature_end",
        "temperature_start",
        "time_step",
        "updated_settings",
        "vacancy_mode",
        "validator_kwargs",
        "validator_name",
        "workflow_base",
        "write_summary_files",
    ]


@pytest.mark.django_db
def test_load_results_from_directories(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="",  # copy over the entire folder
    )

    load_results_from_directories(base_directory=tmp_path)



# -*- coding: utf-8 -*-
from simmate.apps.vasp.inputs.potcar_mappings import PBE_POTCAR_MAPPINGS

HSE_POTCAR_MAPPINGS = PBE_POTCAR_MAPPINGS.copy()
HSE_POTCAR_MAPPINGS.update(
    {
        "Dy": "Dy",
        "Er": "Er",
        "Ho": "Ho",
        "Nd": "Nd",
        "Sm": "Sm",
        "Tb": "Tb",
    }
)
